import sys
import os
import random
import pickle 

from tqdm import tqdm
import numpy as np

import qrem.common.providers.ibmutils.data_converters
from qrem.common import io

#directory_QREM = os.environ["QREM"] +'\\src\\qrem\\'
#sys.path.append(os.path.dirname(directory_QREM))
#os.environ["QREM"] = "/home/tuzjan/Documents/QREM_DEVELOPMENT/QREM_SECRET_DEVELOPMENT/src/qrem/"

import qrem.functions_qrem.ancillary_functions  as anf
import qrem.functions_qrem.functions_data_analysis as fdt 

from qrem.noise_characterization.tomography_design.overlapping.DOTMarginalsAnalyzer import DOTMarginalsAnalyzer  
from qrem.noise_characterization.base_classes.OverlappingTomographyBase import OverlappingTomographyBase
from qrem.noise_characterization.tomography_design.overlapping.SeparableCircuitsCreator import SeparableCircuitsCreator
from qrem.noise_characterization.data_analysis.InitialNoiseAnalyzer import InitialNoiseAnalyzer
from typing import Dict, Tuple, List

from qrem.backends_support.qiskit import qiskit_utilities
from qrem.functions_qrem import functions_data_analysis as fdt

from qrem.noise_simulation.CN.noise_implementation import  simulate_noise_results_dictionary
from qrem.noise_model_generation.CN.NoiseModelGenerator import NoiseModelGenerator

import qrem.common.math as qrem_math
from qrem.common.printer import qprint, qprint_array
from qrem.functions_qrem import functions_benchmarks as fun_ben
from qrem.types import CNModelData
from qrem.cn import mitigation as cnmitigation
from qrem.cn import simulation as cnsimulation 
from datetime import date

from tqdm import tqdm


import statistics 

# Function used to divide a dictionary into two parts. Here it is used to create characterization and benchmark data 

def divide_dictionary(dictionary, items_number, start_index=0):
    counter=0
    divided_dictionary_1 ={}
    divided_dictionary_2={}
    for key,entry in  dictionary.items():
        if counter<start_index+items_number-1:
            divided_dictionary_1[key] = entry
        else:
            divided_dictionary_2[key]=entry    
        counter=counter+1
    return divided_dictionary_1, divided_dictionary_2

def compute_mitigation_errors(mitigated_energy_dictionary, hamiltonian_energy_dictionary,number_of_qubits):
    
    noise_models_mitigated_energy_dictionary_error = {}
    
    for key, benchmark_energy_dictionary in mitigated_energy_dictionary.items():
    
        mitigated_energy_dictionary_error ={}
    
        for index, energy_value in benchmark_energy_dictionary.items():
    
            if index in hamiltonian_energy_dictionary.keys():
    
                mitigated_energy_dictionary_error[index] = abs(energy_value- hamiltonian_energy_dictionary[index]['energy_ideal'])/number_of_qubits
     
        noise_models_mitigated_energy_dictionary_error[key] = mitigated_energy_dictionary_error 
    
    return noise_models_mitigated_energy_dictionary_error

"""

"""
def compute_mitigation_error_median_mean(mitigation_errors_dictionary:Dict,print_results:bool = True):
    benchmark_results_mean_median_dictionary = {'median':{}, 'mean':{}}


    for cluster_assigment, benchmark_results_dictionary in mitigation_errors_dictionary .items():
        benchmark_results_list = list(benchmark_results_dictionary.values()) 
        
    
        benchmark_results_mean_median_dictionary['median'][cluster_assigment] = statistics.median(benchmark_results_list)
        benchmark_results_mean_median_dictionary['mean'][cluster_assigment] = statistics.mean(benchmark_results_list)
    
        if print_results:    
            print(cluster_assigment)
            print('Mitigation error mean' ,benchmark_results_mean_median_dictionary['mean'][cluster_assigment] )
        
            print('Mitigation error median', benchmark_results_mean_median_dictionary['median'][cluster_assigment])

    return benchmark_results_mean_median_dictionary


def estimate_mitigated_energy_for_hamiltonian(results_dictionary:dict[str,dict[str,int]], hamiltonian_dictionary:dict[str,dict], noise_model:type[CNModelData], ensure_proper_probability_distribution:bool = True, product_mitigation:bool =False, return_marginals:bool = False):
    
    """
    Function performs error mitigation based on CN noise model for a particular Hamiltonain and measurement statistics. It is used in usecases and benchmarks. It calls routines from cn/mitigation module.

    Parameters
    ----------
    results_dictionary: dictionary (key - string, value - dictionary)
        Dictionary of experimental results to be mitigated 

     hamiltonian_dictionary: dictionary 
        Dictionary storing hamiltonian data, in this function only weights_dictionary is used 
    
    noise_model: instance of CNModelData class
        Stores information about noise model used in mitigation

    ensure_proper_probability_distribution: bool (optional, by default set to True)
        Controls whether in mitigation one uses proper probability distributions or possibly pseudo probability distributions 
    
    product_mitigation: bool (optional)
        Used for debugging purposes, probably should be removed for final release

    return marginals: bool
         Used for debugging purposes, probably should be removed for final release   
      
        

    """
    
    
    weights_list=list(hamiltonian_dictionary['weights_dictionary'].keys())
    #mitigation is performed 
    
    if not product_mitigation:
        mitigated_marginals = cnmitigation.mitigate_marginals(marginals_list=weights_list,results_dictionary=results_dictionary,noise_model=noise_model, ensure_proper_probability_distribution=ensure_proper_probability_distribution,state_independent_mitigation=False)
    
    else:
        mitigated_marginals = cnmitigation.mitigate_marginals_product(marginals_list=weights_list,results_dictionary=results_dictionary,noise_model=noise_model, ensure_proper_probability_distribution=ensure_proper_probability_distribution)


    
    energy_corrected = \
                    fdt.estimate_energy_from_marginals(weights_dictionary=hamiltonian_dictionary['weights_dictionary'],
                                                       marginals_dictionary= mitigated_marginals[next(iter(mitigated_marginals))])
    
    return {'corrected_energy':energy_corrected} if not return_marginals else {'corrected_energy':energy_corrected, 'mitigated_marginals':mitigated_marginals}

def estimate_mitigated_energy_for_hamiltonians_set(results_dictionary:dict[str,dict[str,int]], hamiltonians_dictionary:dict[str,dict] ,noise_model:type[CNModelData] ,ensure_proper_probability_distribution:bool = True,product_mitigation:bool=False, return_marginals:bool = False):

    """
    Function performs error mitigation based on CN noise model for a set Hamiltonains and set of measurement results. It is used in usecases and benchmarks. 
    Due to QREM convention mitigated measurement statistics from one input state are used to estimate energy of one hamiltonian.
    Function calls estimate_mitigated_energy_for_hamiltonian routine.

    Parameters
    ----------
    results_dictionary: dictionary (key - string, value - dictionary)
        Dictionary of experimental results to be mitigated 

     hamiltonians_dictionary: dictionary 
        Dictionary storing hamiltonians datad 
    
    noise_model: instance of CNModelData class
        Stores information about noise model used in mitigation

    ensure_proper_probability_distribution: bool (optional, by default set to True)
        Controls whether in mitigation one uses proper probability distributions or possibly pseudo probability distributions 
    
    product_mitigation: bool (optional)
        Used for debugging purposes, probably should be removed for final release

    return marginals: bool
         Used for debugging purposes, probably should be removed for final release   
      
        

    """

    mitigation_results_dictionary ={}
    temporary_mitigation_results_dictionary = {}

    mitigation_results_dictionary[ 'corrected_energy']= {hamiltonian_index : 0 for hamiltonian_index in  hamiltonians_dictionary.keys()} 
    if return_marginals:
        mitigation_results_dictionary['mitigated_marginals'] = {hamiltonian_index : 0 for hamiltonian_index in  hamiltonians_dictionary.keys()} 
  

    for state, hamiltoanian_index in tqdm(zip(results_dictionary.keys(), hamiltonians_dictionary.keys()),total=len(list(results_dictionary.keys()))):
        
       
        temporary_mitigation_results_dictionary = estimate_mitigated_energy_for_hamiltonian(results_dictionary={state:results_dictionary[state]},hamiltonian_dictionary= hamiltonians_dictionary[hamiltoanian_index],noise_model=noise_model,ensure_proper_probability_distribution=ensure_proper_probability_distribution,product_mitigation=product_mitigation, return_marginals=return_marginals ) 
        
        mitigation_results_dictionary['corrected_energy'][hamiltoanian_index]  = temporary_mitigation_results_dictionary['corrected_energy']
        
        if return_marginals:
            mitigation_results_dictionary['mitigated_marginals'][hamiltoanian_index]  = temporary_mitigation_results_dictionary['mitigated_marginals']

   
       
    return mitigation_results_dictionary 


def estimate_mitigated_energy_over_noise_models(results_dictionary,hamiltonians_dictionary,noise_models_list,ensure_proper_probability_distribution = True,product_mitigation=False, return_marginals = False ):

    noise_models_mitigation_results_dictionary = {}

    for noise_model in noise_models_list:

        noise_models_mitigation_results_dictionary[noise_model.clusters_tuple] = estimate_mitigated_energy_for_hamiltonians_set(results_dictionary=results_dictionary,hamiltonians_dictionary=hamiltonians_dictionary,noise_model=noise_model, ensure_proper_probability_distribution=ensure_proper_probability_distribution,product_mitigation=product_mitigation, return_marginals= return_marginals)
        
        
         
        #if not return_marginals:
        #    noise_models_mitigated_energy_dictionary[noise_model.clusters_tuple] = estimate_mitigated_energy_for_hamiltonians_set(results_dictionary=results_dictionary,hamiltonians_dictionary=hamiltonians_dictionary,noise_model=noise_model, ensure_proper_probability_distribution=ensure_proper_probability_distribution,product_mitigation=product_mitigation)
            
        #else:
        #  noise_models_mitigated_energy_dictionary[noise_model.clusters_tuple], noise_models_mitigated_marginals_dictionary[noise_model.clusters_tuple] = estimate_mitigated_energy_for_hamiltonians_set(results_dictionary=results_dictionary,hamiltonians_dictionary=hamiltonians_dictionary,noise_model=noise_model, ensure_proper_probability_distribution=ensure_proper_probability_distribution,product_mitigation=product_mitigation,return_marginals=True)
                      
    #structure of the results dictionary is changed 
    noise_models_mitigated_energy_dictionary ={}
    noise_models_marginals_dictionary = {}
    results_dictionary = {}
    
    for key, entry in noise_models_mitigation_results_dictionary.items():
        noise_models_mitigated_energy_dictionary[key] = entry['corrected_energy']
        if return_marginals:
            noise_models_marginals_dictionary[key] = entry['mitigated_marginals']
    
    results_dictionary['corrected_energy'] = noise_models_mitigated_energy_dictionary
    if return_marginals:
        results_dictionary['mitigated_marginals'] = noise_models_marginals_dictionary


    
    return results_dictionary


def compute_noisy_energy_over_noise_models(results_dictionary,hamiltonians_dictionary,noise_models_list):

    noise_models_prediction_results_dictionary = {}

    for noise_model in noise_models_list:

        noise_models_prediction_results_dictionary[noise_model.clusters_tuple] = compute_noisy_energy_for_hamiltonians_set(results_dictionary=results_dictionary,hamiltonians_dictionary=hamiltonians_dictionary,noise_model=noise_model)
        
        
         
   
    noise_models_predicted_energy_dictionary ={}
   
    results_dictionary = {}
    
    for key, entry in noise_models_prediction_results_dictionary.items():
        noise_models_predicted_energy_dictionary[key] = entry['predicted_energy']
    
    
    results_dictionary['predicted_energy'] = noise_models_predicted_energy_dictionary
 


    
    return results_dictionary


def compute_noisy_energy_for_hamiltonians_set(results_dictionary:dict[str,dict[str,int]], hamiltonians_dictionary:dict[str,dict] ,noise_model:type[CNModelData] ):

  

   

    prediction_results_dictionary = {}
  
    prediction_results_dictionary[ 'predicted_energy']= {hamiltonian_index : 0 for hamiltonian_index in  hamiltonians_dictionary.keys()} 
    for state, hamiltoanian_index in tqdm(zip(results_dictionary.keys(), hamiltonians_dictionary.keys()),total=len(list(results_dictionary.keys()))):
        
       
       
        prediction_results_dictionary['predicted_energy'][hamiltoanian_index]  =  compute_noise_energy_for_hamiltonian(hamiltonian_dictionary= hamiltonians_dictionary[hamiltoanian_index],noise_model=noise_model,input_state=state ) 
        
    

   
       
    return prediction_results_dictionary 



def compute_noise_energy_for_hamiltonian( hamiltonian_dictionary:dict[str,dict], noise_model:type[CNModelData],input_state):
    
  
    
    
    weights_dictionary=hamiltonian_dictionary['weights_dictionary']
    #mitigation is performed 
    
    predicted_energy = compute_noisy_energy_prediction(noise_model=noise_model,weights_dictionary=weights_dictionary,input_state=input_state)
    
    return {'predicted_energy':predicted_energy} 



def compute_noisy_energy_prediction(noise_model: type[CNModelData],weights_dictionary:Dict,input_state):
        
       

        needed_pairs = [x for x in weights_dictionary.keys() if len(x) == 2]
        energies_dictionary_now={}
        errors_dictionary_now = {}
       

        noise_matrices_dictionary_model = {cluster: noise_model.noise_matrices[cluster]['averaged']
                                                   for cluster in list(noise_model.clusters_neighborhoods.keys())}

        predicted_energy = fdt.get_noisy_energy_product_noise_model(input_state=input_state,
                                                                            noise_matrices_dictionary=noise_matrices_dictionary_model,
                                                                            needed_pairs=needed_pairs,
                                                                            weights_dictionary_tuples=weights_dictionary)

        
        
        return predicted_energy


def simulate_noisy_experiment( noise_model: type[CNModelData], number_of_circuits: int, number_of_shots: int, number_of_benchmark_circuits:int, data_directory: str ,name_id: str = '',save_data: bool = False,return_ideal_experiment_data: bool = False):

    number_of_qubits = noise_model.number_of_qubits  
    qubit_indices = [i for i in range(number_of_qubits)]

 

    name_string = str(date.today()) + " q" + str(number_of_qubits)+name_id 
    
    OT_creator = OverlappingTomographyBase(number_of_qubits=number_of_qubits, experiment_name='DDOT') 
                                       

    circuits_QDOT = OT_creator.get_random_circuits_list(number_of_circuits=number_of_circuits)
        

    

    circuits_creator = SeparableCircuitsCreator(SDK_name='qiskit', 
                                            qubit_indices=qubit_indices,
                                            descriptions_of_circuits=circuits_QDOT,
                                            experiment_name='DDOT')

    OT_circuits_list = circuits_creator.get_circuits()

    batches = anf.create_batches(circuits_list=OT_circuits_list, circuits_per_job = 300)

    jobs_list = qiskit_utilities.run_batches(backend_name='qasm_simulator',batches=batches, shots=number_of_shots)

    unprocessed_results = qrem.common.providers.ibmutils.data_converters.get_counts_from_qiskit_jobs(jobs_list=jobs_list)

    processed_results = fdt.convert_counts_overlapping_tomography(counts_dictionary=unprocessed_results, experiment_name='DDOT')


    file_name_results  = 'DDOT_simulation_workflow_results_' + name_string +'.pkl'
    io.save(dictionary_to_save=processed_results,
                                directory=data_directory,
                                custom_filename=file_name_results)




    results_dictionary =  {}
    if return_ideal_experiment_data:
        
        marginals_analyzer_ideal = DOTMarginalsAnalyzer(results_dictionary_ddot=processed_results)

        single_qubits = [(i,) for i in range(number_of_qubits)]
        pairs_of_qubits = [(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)]

        marginals_to_mitigate=list(noise_model.clusters_tuple)  + single_qubits + pairs_of_qubits

        marginals_analyzer_ideal.compute_all_marginals(subsets_dictionary=marginals_to_mitigate,show_progress_bar=True,multiprocessing=True)
        ideal_experiments_dic, ideal_benchmarks_dic = divide_dictionary(marginals_analyzer_ideal.marginals_dictionary,len(marginals_analyzer_ideal.marginals_dictionary.keys())-number_of_benchmark_circuits)
        results_dictionary['ideal_experiments_dictionary'] = ideal_experiments_dic
        results_dictionary['ideal_benchmarks_dictionary'] = ideal_benchmarks_dic



    noisy_results_dictionary=cnsimulation.simulate_noise_results_dictionary(processed_results,noise_model)

    results_dictionary['noisy_results_dictionary'] = noisy_results_dictionary



    if save_data:
        file_name_noisy_results  = 'DDOT_noisy_simulation_results_' + name_string +'.pkl'

        io.save(dictionary_to_save=results_dictionary,
                                directory=data_directory,
                                custom_filename=file_name_noisy_results)
    return results_dictionary

    
     

    print('Simulation of noise is finished')

def create_data_to_test_mitigation(number_of_qubits, noise_model, number_of_circuits, number_of_shots,number_of_benchmark_circuits,data_directory,return_old_mitigation_data = False):
    

    number_of_benchmark_circuits = number_of_benchmark_circuits -1


    


    qubit_indices = [i for i in range(number_of_qubits)]

    single_qubits = [(i,) for i in range(number_of_qubits)]
    pairs_of_qubits = [(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)]

    marginals_to_mitigate=list(noise_model.clusters_tuple)  + single_qubits + pairs_of_qubits

    name_string = str(date.today()) + " q" + str(number_of_qubits) 
    
    OT_creator = OverlappingTomographyBase(number_of_qubits=number_of_qubits, experiment_name='DDOT') 
                                       

    circuits_QDOT = OT_creator.get_random_circuits_list(number_of_circuits=number_of_circuits)
        

    

    circuits_creator = SeparableCircuitsCreator(SDK_name='qiskit', 
                                            qubit_indices=qubit_indices,
                                            descriptions_of_circuits=circuits_QDOT,
                                            experiment_name='DDOT')

    OT_circuits_list = circuits_creator.get_circuits()

    batches = anf.create_batches(circuits_list=OT_circuits_list, circuits_per_job = 300)

    jobs_list = qiskit_utilities.run_batches(backend_name='qasm_simulator',batches=batches, shots=number_of_shots)

    unprocessed_results = qrem.common.providers.ibmutils.data_converters.get_counts_from_qiskit_jobs(jobs_list=jobs_list)

    processed_results = fdt.convert_counts_overlapping_tomography(counts_dictionary=unprocessed_results, experiment_name='DDOT')


    file_name_results  = 'DDOT_simulation_workflow_results_' + name_string +'.pkl'
    io.save(dictionary_to_save=processed_results,
                                directory=data_directory,
                                custom_filename=file_name_results)


    marginals_analyzer_ideal = DOTMarginalsAnalyzer(results_dictionary_ddot=processed_results)

    marginals_analyzer_ideal.compute_all_marginals(subsets_dictionary=marginals_to_mitigate,show_progress_bar=True,multiprocessing=False)

    ideal_experiments_dic, ideal_benchmarks_dic = divide_dictionary(marginals_analyzer_ideal.marginals_dictionary,len(marginals_analyzer_ideal.marginals_dictionary.keys())-number_of_benchmark_circuits)



    noisy_results_dictionary=cnsimulation.simulate_noise_results_dictionary(processed_results,noise_model)


    file_name_noisy_results  = 'DDOT_noisy_simulation_results_' + name_string +'.pkl'

    io.save(dictionary_to_save=noisy_results_dictionary,
                                directory=data_directory,
                                custom_filename=file_name_noisy_results)

    


    io.save(dictionary_to_save=noisy_results_dictionary,
                                directory=data_directory,
                                custom_filename=file_name_noisy_results)

    print('Simulation of noise is finished')

    locality=2
    subset_of_qubits = qrem_math.get_k_local_subsets(number_of_qubits, 2,True)
    #subset_of_qubits = pairs_of_qubits
    #subset_of_qubits +=  [(i,) for i in range(number_of_qubits)]





    print('Marginals calculation starts')

    marginals_analyzer = DOTMarginalsAnalyzer(results_dictionary_ddot=noisy_results_dictionary)


    marginals_analyzer.compute_all_marginals(subset_of_qubits,show_progress_bar=True,multiprocessing=False)

    marginals_dictionary=marginals_analyzer.marginals_dictionary

    dictionary_to_save = {'marginals_dictionary':marginals_dictionary}

    characterization_results_dictionary, benchmarks_results_dictionary=divide_dictionary(noisy_results_dictionary,len(noisy_results_dictionary.keys())-number_of_benchmark_circuits)
    characterization_marginals_dictionary, benchmarks_marginals_dictionary = divide_dictionary(marginals_dictionary,len(noisy_results_dictionary.keys())-number_of_benchmark_circuits)


    file_name_marginals = 'DDOT_marginals_workflow_' + name_string +'.pkl'

    io.save(dictionary_to_save=dictionary_to_save,
                                directory=data_directory,
                                custom_filename=file_name_marginals)


    io.save(dictionary_to_save=dictionary_to_save,
                                directory=data_directory,
                                custom_filename=file_name_marginals)

    marginals_analyzer = DOTMarginalsAnalyzer(results_dictionary_ddot=characterization_results_dictionary,marginals_dictionary=characterization_marginals_dictionary)

    marginals_analyzer.compute_subset_noise_matrices_averaged(subset_of_qubits)


    POVMs_dictionary_diag= marginals_analyzer._noise_matrices_dictionary




    file_name_POVMs  = 'DDOT_POVMs_PLS_workflow_' + name_string +'.pkl'


    POVMs_dictionary = {}
    for key, entry in POVMs_dictionary_diag.items():
        list_of_matrices=[]
        for matrix in POVMs_dictionary_diag[key]['averaged']:
            list_of_matrices.append(np.diag(matrix))

        POVMs_dictionary[key]=list_of_matrices

    dictionary_to_save = {'POVMs_dictionary':POVMs_dictionary}


    characterization_POVMs_dictionary=POVMs_dictionary

    noise_analyzer = InitialNoiseAnalyzer(results_dictionary=characterization_results_dictionary,
                                      marginals_dictionary=characterization_marginals_dictionary,
                                      POVM_dictionary=characterization_POVMs_dictionary
                                              )

    io.save(dictionary_to_save=dictionary_to_save,
                                directory=data_directory,
                                custom_filename=file_name_POVMs)


    qubit_indices = list(range(number_of_qubits))

    distances_types_correlations = [('worst_case', 'classical')]
        
    noise_analyzer.compute_correlations_data_pairs(qubit_indices=range(number_of_qubits),
                                              distances_types=distances_types_correlations)

    correlations_data = noise_analyzer.correlations_data

    dictionary_to_save = {'correlations_data':correlations_data,
                      
                     }

    file_name_errors  = 'DDOT_correlations_workflow_' + name_string +'.pkl'
    io.save(dictionary_to_save=dictionary_to_save,
                                directory=data_directory,
                                custom_filename=file_name_errors)



    characterization_noise_matrix_dictionary = fdt.get_noise_matrices_from_POVMs_dictionary(characterization_POVMs_dictionary)


    noise_model_analyzer = NoiseModelGenerator(results_dictionary=characterization_results_dictionary,
                                           marginals_dictionary=characterization_marginals_dictionary,
                                           noise_matrices_dictionary=characterization_noise_matrix_dictionary ,
                                           correlations_data=correlations_data
                                           )



    sizes_clusters = [2,3,4]

    distance_type = 'worst_case'
    correlations_type = 'classical'
    correlations_table = correlations_data[distance_type][correlations_type]

    alpha_hyperparameters = np.linspace(0.0, 3.0, 16)
    alpha_hyperparameters = [np.round(alpha, 3) for alpha in alpha_hyperparameters]
    alpha_hyperparameters=[1]

    all_clusters_sets_dictionary = {tuple([(qi,) for qi in range(number_of_qubits)]):None}
    #all_clusters_sets_dictionary= {}

    for max_cluster_size in sizes_clusters:
        print("\nCurrent max cluster size:", max_cluster_size, 'red')

        for alpha_multiplier in alpha_hyperparameters:
            clustering_function_arguments = {'alpha_multiplier': alpha_multiplier}

            clusters_tuples_now, score = noise_model_analyzer.compute_clusters(correlations_table=correlations_table,
                                                                           maximal_size=max_cluster_size,
                                                                           method_kwargs=clustering_function_arguments)
           # neighborhoods = noise_model_analyzer.find_all_neighborhoods(maximal_size=2)
            if clusters_tuples_now in all_clusters_sets_dictionary.keys():
                all_clusters_sets_dictionary[clusters_tuples_now].append(
                    (max_cluster_size, clustering_function_arguments['alpha_multiplier'], score))#, neighborhoods))
            else:
                all_clusters_sets_dictionary[clusters_tuples_now] = [
                    (max_cluster_size, clustering_function_arguments['alpha_multiplier'], score)]#, neighborhoods)]












    all_clusters_sets_list = list(all_clusters_sets_dictionary.keys())

    cluster_subsets = []

    for cluster_list in all_clusters_sets_list:
        for cluster in cluster_list:
            if cluster not in cluster_subsets:
                cluster_subsets.append(cluster)

    marginals_analyzer.compute_all_marginals(cluster_subsets,
                                         show_progress_bar=True,
                                         multiprocessing=False)

    marginals_dictionary_clusters = marginals_analyzer.marginals_dictionary


    noise_model_list =[]
    noise_models_noise_matrices ={}
    for cluster_list in all_clusters_sets_list:
        #marginals_analyzer.noise_matrices_dictionary = {}
        marginals_analyzer_noise_model = DOTMarginalsAnalyzer(results_dictionary_ddot=   characterization_results_dictionary)
        marginals_analyzer_noise_model.compute_subset_noise_matrices_averaged(subsets_list=cluster_list,show_progress_bar=True)
        #for test purposes
        noise_models_noise_matrices[cluster_list] = marginals_analyzer.noise_matrices_dictionary 
        #marginals_analyzer.compute_subset_noise_matrices_averaged(subsets_list=cluster_list,show_progress_bar=True)
        ##addidtion starts###
        #POVMs_dictionary_diag= marginals_analyzer._noise_matrices_dictionary
        #POVMs_dictionary = {}
        #for key, entry in POVMs_dictionary_diag.items():
        #    list_of_matrices=[]
        #    for matrix in POVMs_dictionary_diag[key]['averaged']:
        #        list_of_matrices.append(np.diag(matrix))

        #    POVMs_dictionary[key]=list_of_matrices

        #noise_matrices_dictionary = fdt.get_noise_matrices_from_POVMs_dictionary(characterization_POVMs_dictionary)
        ###addition ends###
        #new_noise_matrix_dictionary = noise_matrices_dictionary
        noise_matrices_dictionary = marginals_analyzer_noise_model.noise_matrices_dictionary
        new_noise_matrix_dictionary = {}
        for key, noise_matrix in noise_matrices_dictionary.items():
            new_noise_matrix_dictionary[key] =noise_matrix['averaged']
        noise_model = CNModelData(number_of_qubits)
        noise_model.set_clusters_tuple(clusters_tuple=tuple(cluster_list))
        noise_model.set_noise_matrices_dictionary(noise_matrices_dictionary=new_noise_matrix_dictionary)
        noise_model_list.append(noise_model)






    if return_old_mitigation_data:

        marginals_analyzer_noise_model_old = DOTMarginalsAnalyzer(results_dictionary_ddot=   characterization_results_dictionary)
        marginals_analyzer_noise_model_old.compute_subset_noise_matrices_averaged(subsets_list=cluster_subsets,show_progress_bar=True)

        #marginals_analyzer.compute_subset_noise_matrices_averaged(subsets_list=chosen_clustering,show_progress_bar=True)



        #marginals_analyzer.compute_subset_noise_matrices_averaged(chosen_clustering,
        #                                 show_progress_bar=True)
        noise_matrices_dictionary_old = marginals_analyzer_noise_model_old.noise_matrices_dictionary 
        
        correction_matrices, correction_indices,mitigation_data_dictionary = fdt.get_multiple_mitigation_strategies_clusters_for_pairs_of_qubits(
        pairs_of_qubits=pairs_of_qubits,
        clusters_sets=all_clusters_sets_dictionary,
        dictionary_results=characterization_results_dictionary,
        #marginals_dictionary={},
        noise_matrices_dictionary=noise_matrices_dictionary_old,
        show_progress_bar=True)

        marginals_analyzer.noise_matrices_dictionary  = {}
        marginals_analyzer.compute_subset_noise_matrices_averaged(subsets_list= cluster_subsets,show_progress_bar=True)

        return {'new_mitigation_data':[noise_model_list,benchmarks_results_dictionary,benchmarks_marginals_dictionary],'old_mitigation_data': [correction_matrices, correction_indices,mitigation_data_dictionary,    all_clusters_sets_dictionary, noise_matrices_dictionary_old ], 'ideal_benchmarks_dictionary': ideal_benchmarks_dic}
        
    else:
        return noise_model_list,benchmarks_results_dictionary,benchmarks_marginals_dictionary
    


    





import numpy as np
import scipy
from itertools import combinations
from typing import Dict, Tuple, List

from qrem.ctmp.modeltools.qubit_utils import outeq


# order of basis elements
basis_dict = {0 : "00", 1 : "11", 2: "01", 3 : "10"}
# reverse basis dictionary
rbd = {v: k for k, v in basis_dict.items()}

def rates_from_exp_results(exp_results: Dict[str, Dict[str, int]], n: int) -> List[Tuple]:
    """Calibrates the error rates to experimental results, as described in the paper. Return a list of model
    rates in the format as in CTMPModelData class.

    exp_results - dictionary of experimental results in the form {input string: dict of output counts}
    e.g. {"001" : {"000" : 56, "001" : 333, ...}}
    n - number of qubits

    The function assumes that experimental results are complete, i.e. suffice to determine the model rates.
    """
    # todo: check if the input set is complete
    pairs = list(combinations(range(0, n), 2))
    rates = []
    G_dict = {}
    for j, k in pairs:
        A = np.zeros((4,4))
        # not optimized for efficiency
        for v in range(0, 4):
            vs = basis_dict[v]
            v_results = {x:value for x, value in exp_results.items() if x[j]+x[k]==vs }
            filtered_dict = {}
            for x, value_x in v_results.items():
                filtered_values = {y:value_y for y, value_y in value_x.items() if outeq(x, y, j, k) }
                filtered_dict[x] = filtered_values
            total_count = sum({k:sum(val.values()) for k, val in filtered_dict.items()}.values())
            for w in range(0, 4):
                ws = basis_dict[w]
                final_count = 0
                for x, value_x in filtered_dict.items():
                    final_sum = sum([value_y for y, value_y in value_x.items() if y[j]+y[k]==ws])
                    final_count += final_sum
                if total_count > 0:
                    A[w, v] = final_count / total_count
                else:
                    raise Exception("Total count equal zero for j = ", j, "k = ", k, "v = ", vs)
        G = scipy.linalg.logm(A)
        if np.iscomplex(G).any():
            G = np.zeros_like(A)
            print("Warning: matrix log has complex entries, setting rates to 0.")
        G_prime = G.clip(min=0)
        G_dict[(j,k)] = G_prime
        rates.append((j, k, "00","11", G_prime[rbd["11"], rbd["00"]]))
        rates.append((j, k, "11","00", G_prime[rbd["00"], rbd["11"]]))
        rates.append((j, k, "01","10", G_prime[rbd["01"], rbd["10"]]))
        rates.append((j, k, "10","01", G_prime[rbd["10"], rbd["01"]]))
    for j in range(n):
        d = {k:v for k, v in G_dict.items() if j in k}
        # formula is different than in the paper since here always j \leq k
        r01 = 1/(2*(n-1)) *sum([G[rbd["10"], rbd["00"]] + G[rbd["01"], rbd["00"]] +
                                G[rbd["11"], rbd["01"]] + G[rbd["11"], rbd["10"]]
                                for G in d.values()])
        r10 = 1/(2*(n-1)) *sum([G[rbd["00"], rbd["10"]] + G[rbd["00"], rbd["01"]] +
                                G[rbd["01"], rbd["11"]] + G[rbd["10"], rbd["11"]]
                                for G in d.values()])
        rates.append((j, j, "0", "1", r01))
        rates.append((j, j, "1", "0", r10))
    return rates



"""qrem.common.math module contains all mathematical funcitons, 
that are useful and used throughout the qrem package.

"""

import warnings #standard imports go first; kets make it those that are python/system/code -related
import functools as ft
import copy
import itertools
from typing import List, Dict, Optional, Callable, Tuple

import numpy as np #then 3rd party imports
import numpy.typing as npt #then 3rd party imports
import cmath # lets add internal meth related imports here

# then internal old imports

import qrem.common.utils #then internal new imports
from qrem.common.constants import SIGNIFICANT_DIGITS

#TODO_PP >> talk with Paweł some format semantics
class Constants:
    """ Contains static methods that return helpful mathematical constants, matrices, arrays etc.

    Attributes
    ----------
    pauli_dict_test : test for attribute documentation

    Notes
    -----
    The choice to use static methods rather than variables was made for documentation/pylint readability. 
    Before Python 3.11 the way to go would be to wrap @classmethod around @property, so you could access this methods as e.g Constant.pauli_sigmas['id'].
    It is only possible however in Python 3.9-3.10. 
    
    Another possibility would be to use (i) module variable or (ii) class variable, however first one is hard to document to provide
    e.g. nice hints in PyCharm or VSCode, with the second you can do this for example with Attributes section at class-level docstring,
    but it won't give you e.g. nice hover hints in VSCode. 
    For that reason now we need to bear with a bit ugly:
    Constant.pauli_sigmas()['id']

    """

    #(PP) just for testing some documentation stuff
    pauli_dict_test = {1,2,3}

    @staticmethod
    def pauli_sigmas():
        """Returns a dictionary with Paili sigmas: id, I, X, Y, Z; all dictionary values are numpy arrays with complex elements. """
        #(PP) put module level variables into classes to group them and be able to document their purpose. 
        
        return {
            'id': np.array([[1., 0j], [0, 1]]),
            'I':np.array([[1., 0j], [0, 1]]),
            'X': np.array([[0.j, 1],
                        [1, 0]]),
            'Y': np.array([[0., -1j],
                        [1j, 0]]),
            'Z': np.array([[1., 0j],
                        [0, -1]])
        }

    @staticmethod
    def bell_states():
        """Returns a dictionary with Bell states : phi+, phi-; psi+; psi-; all dictionary values are numpy arrays. """


        return {'phi+': 1 / np.sqrt(2) * np.array([1, 0, 0, 1]),
                'phi-': 1 / np.sqrt(2) * np.array([1, 0, 0, -1]),
                'psi+': 1 / np.sqrt(2) * np.array([0, 1, 1, 0]),
                'psi-': 1 / np.sqrt(2) * np.array([0, 1, -1, 0]), }



    @staticmethod
    def standard_gates():
        """Returns a dictionary with Standard Gates : 1,I,id,X,Y,Z,S,T,H,SWAP; all dictionary values are numpy complex arrays. """
        #MOVE TO >> common.quantum.QConstans in the future

        return {'1': np.array([[1., 0j], [0, 1]], dtype=complex),
                'I': np.array([[1., 0j], [0, 1]], dtype=complex),
                'id': np.array([[1., 0j], [0, 1]], dtype=complex),
                'X': np.array([[0.j, 1], [1, 0]], dtype=complex),
                'Y': np.array([[0., -1j], [1j, 0]], dtype=complex),
                'Z': np.array([[1., 0j], [0, -1]], dtype=complex),
                'S': np.array([[1, 0], [0, 1j]], dtype=complex),
                'T': np.array([[1, 0], [0, np.exp(1j * np.pi / 4)]], dtype=complex),
                'H': 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]], dtype=complex),
                'SWAP': np.array(  [[1, 0, 0, 0],
                                    [0, 0, 1, 0],
                                    [0, 1, 0, 0],
                                    [0, 0, 0, 1]], dtype=complex)}

    @staticmethod
    def pauli_eigenkets():
        """Returns a dictionary with Pauli Eigenkets : x+,x-,y+,y-,z+,zi; all dictionary values are numpy complex arrays. """
        #MOVE TO >> common.quantum.QConstans in the future ?

        return {'x+': 1 / np.sqrt(2) * np.array([[1], [1]], dtype=complex),
                       'x-': 1 / np.sqrt(2) * np.array([[1], [-1]], dtype=complex),
                       'y+': 1 / np.sqrt(2) * np.array([[1], [1j]], dtype=complex),
                       'y-': 1 / np.sqrt(2) * np.array([[1], [-1j]], dtype=complex),
                       'z+': np.array([[1], [0]], dtype=complex),
                       'z-': np.array([[0], [1]], dtype=complex)}

    @staticmethod
    def pauli_eigenkets_listed(order="ZXY"):
        """Returns a list with Pauli Eigenkets, Default order z+z-x+x-y+y-. Order is defined by a strin, e.g. "ZYX or "XYZ" """
        #MOVE TO >> common.quantum.QConstans in the future ?
        output_list = []
        for el in order:
            if el =="Z":
                output_list.append(Constants.pauli_eigenkets("z+"))
                output_list.append(Constants.pauli_eigenkets("z-"))
            if el =="Y":
                output_list.append(Constants.pauli_eigenkets("y+"))
                output_list.append(Constants.pauli_eigenkets("y-"))
            if el =="X":
                output_list.append(Constants.pauli_eigenkets("x+"))
                output_list.append(Constants.pauli_eigenkets("x-"))
            else:
                raise ValueError(f"Wrong label in the order definition: '{order}' - a letter was not recognized.")

        return output_list

                       
    @staticmethod
    def pauli_measurements():
        """Returns a dictionary with Pauli Measurments on 'X', 'Y', 'Z'. """
        #MOVE TO >> common.quantum.QConstans in the future 

        return {'X': [ketbra(Constants.pauli_eigenkets()['x+']),
                        ketbra(Constants.pauli_eigenkets()['x-'])],
                'Y': [ketbra(Constants.pauli_eigenket()['y+']),
                        ketbra(Constants.pauli_eigenkets()['y-'])],
                'Z': [ketbra(Constants.pauli_eigenkets()['z+']),
                        ketbra(Constants.pauli_eigenkets()['z-'])],
                }

# ===================================================
# Helpful Vector/matrix math
# ===================================================

def outer_product(ket1, ket2=None):
    """Outer product of two vectors.

    Notes
    -----
    - PP: should it just be np.outer(ket1, ket2)?
    - MO: what about complex entries?  
    """

    if ket2 is None:
        ket2 = ket1
    return ket1 @ np.matrix.getH(ket2)
    
def ketbra(ket):
    """Ketbra operation. Can be simplified with numpy
    """

    return outer_product(ket1=ket, ket2=None)

def matrix_zero_check(potential_zero_matrix: np.ndarray,
               significant_digits=SIGNIFICANT_DIGITS) -> bool:
    """
    Functions that checks if matrix is zero. Wraps numpy allclose.

    Parameters
    ----------
    potential_zero_matrix : NDArray
    significant_digits : int

    Returns
    -------
    bool
        True if matrix is 0 with given significant digits accuracy
    """
    threshold = 10 ** (-significant_digits)
    size = list(potential_zero_matrix.shape)

    zeros = np.zeros(size, dtype=type(potential_zero_matrix[0, 0]))
    return np.allclose(potential_zero_matrix, zeros, rtol=threshold)

def identity_check(m_a, significant_digits=SIGNIFICANT_DIGITS) -> bool:
    """Function that checks if matrix is identity (up to global phase), with significant_digits accuracy. 

    Returns
    -------
    bool
        If matrix is identity  returns True
    """
    size = np.size(m_a, 0)
    m_b = round_array_to_ndigits(m_a)

    m_b_phase = cmath.phase(m_b[0, 0])
    m_b_prime = cmath.exp(-1j * m_b_phase) * m_b

    identity_matrix = np.identity(size)

    checking_zeros = round_array_to_ndigits(m_b_prime - identity_matrix)

    return True if matrix_zero_check(checking_zeros, significant_digits) else False

#TEST_ME - old version changed
def round_array_to_ndigits(m_a: np.ndarray, significant_digits=SIGNIFICANT_DIGITS) -> np.ndarray:
    """perform round operation on a given array m_a.

    Notes
    -----
    - PP: refactored from some very werid construction. Should have the same functionality. 
    """
    # OLD VERSION
    
    # # TODO PP: should be: np.around(m_a,  decimal)
    # # TODO TR: Comment this method_name, and intention behind it.
    # m_b = np.array(copy.deepcopy(m_a))
    # with np.nditer(m_b, op_flags=['readwrite']) as it:
    #     for x in it:
    #         x[...] = np.round(x, decimal)

    # return m_b
    return np.around(m_a,  decimals=significant_digits)

#TEST_ME e.g. if m_a has different size then pauli sigmas
def is_pauli_x(m_a) -> bool:
    """Is a matrix an X Pauli Sigma? """
    return True if matrix_zero_check(m_a - Constants.pauli_sigmas()['X']) else False

# TEST_ME Not sure why the negative entries are being cut - check if this is an isue somewhere (MO)
def is_matrix_stochastic(potentially_stochastic_matrix: np.ndarray,
                        stochasticity_type: Optional[str] = 'left',
                        significant_digits: Optional[float] = SIGNIFICANT_DIGITS) -> bool:
    """Test stochasticity of a given input matrix: left|right|ortho. Default is left. 
    You can specify significant digits of the check, default is 9 (accuracy 10^9).

    Parameters
    ----------
    potentially_stochastic_matrix : NDArray 
    stochasticity_type: {'left', 'right', 'ortho', 'doubly'}
        string specyfing what type of stochasticity we want to tests; otrho option is equal to doubly
    significant_digits, default=9
        significant digits specifying the accuracy to which entries need to be positive rows/ columns

    Returns
    -------
        bool 
            is matrix stochastic 

    Raises
    ------
        ValueError
            Wrong stochasticity_type of stochasticity
    
    Notes
    -----
    
    """
    accuracy = 10 ** (-significant_digits)
    shape = potentially_stochastic_matrix.shape[0]

    potentially_stochastic_matrix2 = copy.deepcopy(potentially_stochastic_matrix)
    potentially_stochastic_matrix2[abs(potentially_stochastic_matrix2) < accuracy] = 0
    if np.any(potentially_stochastic_matrix2 < 0):
        return False

    if stochasticity_type == 'left':
        for index_row in range(shape):
            one_now = sum(potentially_stochastic_matrix[:, index_row])

            if abs(1 - one_now) >= accuracy:
                return False
    elif stochasticity_type == 'right':
        for index_row in range(shape):
            one_now = sum(potentially_stochastic_matrix[index_row, :])

            if abs(1 - one_now) >= accuracy:
                return False
    elif stochasticity_type == 'ortho' or stochasticity_type == 'doubly':
        for index_both in range(shape):
            one_now = sum(potentially_stochastic_matrix[:, index_both])
            one_now2 = sum(potentially_stochastic_matrix[index_both, :])
            if abs(1 - one_now) >= accuracy or abs(1 - one_now2) >= accuracy:
                return False

    else:
        raise ValueError('Wrong stochasticity_type of stochasticity')

    return True

def kronecker_product(arguments: List[npt.ArrayLike]) -> npt.NDArray:
    """applies kronecker product to all arrays in the arguments list"""
    return ft.reduce(np.kron, arguments)


# ===================================================
# Helpful Quantum-related
# ===================================================

def apply_unitary_channel(matrix_to_be_rotated, unitary_operator) ->np.ndarray:
    """Apply unitary channel to a matrix: unitary_operator @ matrix_to_be_rotated @ np.matrix.getH(unitary_operator)
    """
    return unitary_operator @ matrix_to_be_rotated @ np.matrix.getH(unitary_operator)

def get_k_local_subsets(number_of_elements: int, subset_size: int, all_sizes_up_to_k = False) -> List:
    """Create all subsets of size (locality) k=subset_size  of n=number_of_elements elements set.
    If all_sizes_up_to_k is set to True, will return all subsets of size from 1 up to k.

    Parameters
    ----------
    number_of_elements: int
        number of qubits
    subset_size: int
        max number of element in a subset
    all_sizes_up_to_k: bool
        should all sizes from 1 up to k be taken into consideration
    Returns
    -------
    List
        list of all subsets up to k=subset_size number of elements
    """
    output = list(itertools.combinations(range(number_of_elements), subset_size))
    
    if not all_sizes_up_to_k:
        return output
    else:
        for i in range(subset_size-1):
            output += list(itertools.combinations(range(number_of_elements), i + 1))
        return output



# ===================================================
# Helpful functions - probability distributions
# ===================================================
def get_sample_from_multinomial_distribution(probability_distribution: npt.ArrayLike, seed=None,
                                                    method='numpy') -> int:
    """Draw one single sample from multinomial distribution (numpy implementation only for now)

    Parameters
    ----------
    probability_distribution: List(int)
        list of probability distribution values
    method: str
        method of drawing, currently based on numpy random and multinomial functions
    seed: int
        seed to initialize numpy random. Default is None

    Returns
    -------
    int
        single sample
    """

    if method.upper() == 'NUMPY':
        rng = np.random.default_rng(seed)
        samples = rng.multinomial(1, probability_distribution)
    else:
        raise ValueError("wrong sampling method")

    return np.argmax(samples)

# ===================================================
# Helpful functions - TVD calculation
# ===================================================



def compute_TVD(p:np.ndarray,q:np.ndarray):
    """
    Calculation of TVD between two probability distributions. 
    """
    return  0.5*np.linalg.norm(p.flatten()-q.flatten(),ord=1)



# ===================================================
# Helpful functions - matrix/vector SWAP 
# ===================================================


def permute_composite_matrix(qubits_list,noise_matrix):
    qubits_list_sorted = copy.copy(qubits_list)
    noise_matrix_permuted = copy.copy(noise_matrix)
    
    qubits_list_sorted.sort()
    while qubits_list != qubits_list_sorted:
        for index_qubit in range(len(qubits_list)-1):
            if qubits_list[index_qubit+1]<qubits_list[index_qubit]:
                noise_matrix_permuted = permute_matrix(noise_matrix_permuted,len(qubits_list),[index_qubit+1,index_qubit+2])
                qubits_list[index_qubit], qubits_list[index_qubit+1] = qubits_list[index_qubit+1], qubits_list[index_qubit]  
    return noise_matrix_permuted

def permute_composite_vector(qubits_list,vector):
    qubits_list_sorted = copy.copy(qubits_list)
    vector_permuted = copy.copy(vector)
    
    qubits_list_sorted.sort()
    while qubits_list != qubits_list_sorted:
        for index_qubit in range(len(qubits_list)-1):
            if qubits_list[index_qubit+1]<qubits_list[index_qubit]:
                vector_permuted= permute_vector(vector_permuted,len(qubits_list),[index_qubit+1,index_qubit+2])
                qubits_list[index_qubit], qubits_list[index_qubit+1] = qubits_list[index_qubit+1], qubits_list[index_qubit]  
    return vector_permuted


def permute_vector(vector, n, transposition):
    # Swap qubits (subspaces) in 2**number_of_qubits dimensional matrix
    # number_of_qubits - number of qubits
    # transposition - which qubits to SWAP.
    # IMPORTANT: in transposition they are numbered from 1

    swap = qubit_swap(n, transposition)
    return swap @ vector


def permute_matrix(matrix, n, transposition):
    # Swap qubits (subspaces) in 2**number_of_qubits dimensional matrix
    # number_of_qubits - number of qubits
    # transposition - which qubits to SWAP.
    # IMPORTANT: in transposition they are numbered from 1

    swap = qubit_swap(n, transposition)
    return swap @ matrix @ swap


def qubit_swap(n, transposition=(1, 1)):
    # create swap between two qubits in 2**number_of_qubits dimensional space
    # labels inside transpositions start from 1.

    D = 2 ** n
    # renumerate for Python convention
    i, j = transposition[0] - 1, transposition[1] - 1

    names = [(bin(j)[2:].zfill(n)) for j in list(range(2 ** n))]

    new_names = copy.copy(names)
    # exchange classical register bits with labels in transposition
    # this defines new order in classical register which respects qubit SWAP
    for k in range(len(new_names)):
        string = list(new_names[k])
        string[i], string[j] = string[j], string[i]
        new_names[k] = ''.join([s for s in string])

    transformation = np.eye(D)

    for x in range(len(names)):
        bit = int(new_names[x], 2)

        # if we need to change the bit, let's do this
        if bit != x:
            transformation[x, x] = 0
            transformation[bit, bit] = 0
            transformation[bit, x] = 1
            transformation[x, bit] = 1

    return transformation


"""
Created on 01.03.2021

@author: Oskar Słowik
@contact: osslowik@gmail.com
"""

import copy
import math
import random
import numpy as np

from tqdm import tqdm

from qrem.noise_visualization.functions_data_visualization import *


class InfinityException(Exception):
    """Class for handling infinity"""
    pass


class ClusterSizeError(NameError):
    """Class for handling too small max cluster size"""
    pass

class RejectionException(Exception):
    """ Class for handling rejection of all operations """
    pass


# TODO OS: allow 2 qubit clusters but in this case skip the rest of algorithms since init is optimal
# cluster size function.
def f_clust_sharp(C_size, C_maxsize):
    # if(C_max<3):
    #    raise ClusterSizeError
    val = 0
    if C_size > C_maxsize:
        raise InfinityException
    return val


# cluster size function. Assumes C_max>=3. If not, raises exception.
def f_clust(C_size, C_maxsize):
    # if(C_max<3):
    #    raise ClusterSizeError
    val = 0
    if C_size > C_maxsize:
        raise InfinityException
    elif C_size < 3:
        val = 0
    else:
        val = math.pow(C_size / C_maxsize, 2)
    return val


# cluster size function. Assumes C_max>=3. If not, raises exception.
def f_clust_square_sharp(C_size, C_maxsize):
    # if(C_max<3):
    #    raise ClusterSizeError
    val = 0
    if C_size > C_maxsize:
        raise InfinityException
    else:
        val = math.pow(C_size, 2)
    return val


# average inter-cluster attractivness function S_{i,j}
def get_S_ij_av(C_i, C_j, correlations_table):
    val = 0
    for k in C_i:
        for l in C_j:
            val = val + (correlations_table[k, l] + correlations_table[l, k]) / (2 * len(C_i) * len(C_j))
    return val


# intra-cluster attractivness function S_{i}
def get_S_i(C_i, correlations_table):
    if len(C_i) < 2:
        return 0
    val = 0
    for k in C_i:
        for l in C_i:
            if (k == l):
                continue
            val = val + correlations_table[k, l] / (len(C_i) * (len(C_i) - 1))
    return val


# intra-cluster cumulative attractivness function Sc_{i}
def get_Sc_i(C_i, correlations_table):
    if len(C_i) < 2:
        return 0
    val = 0
    for k in C_i:
        for l in C_i:
            if (k == l):
                continue
            val = val + correlations_table[k, l]
    return val


def cost_function_simple_cummulative(partition, correlations_table, alpha, C_maxsize):
    """
    Calculates the value of the objective function phi on partition.

    :param partition: partition
    :type partition: List[int]

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :return value of phi on partition.
    :rtype: float
    """

    no_of_clusters = len(partition)
    val = 0
    for C_i in partition:
        try:
            val = val - alpha * f_clust_square_sharp(len(C_i), C_maxsize)
        except InfinityException:
            raise InfinityException

    for i in range(no_of_clusters):
        C_i = partition[i]
        S_i = get_Sc_i(C_i, correlations_table)
        val = val + S_i
    return val

"""def cost_function_simple_cummulative(partition, correlations_table, alpha, C_maxsize):
    no_of_clusters = len(partition)
    val = 0
    for C_i in partition:
        try:
            val = val - alpha * f_clust_square_sharp(len(C_i), C_maxsize)
        except InfinityException:
            raise InfinityException

    for i in range(no_of_clusters):
        C_i = partition[i]
        S_i = get_Sc_i(C_i, correlations_table)
        val = val + S_i
    return val

"""
# returns the value of cost function (simpler version; no badness) and S table - a symmetric matrix which is defined by S_ij=S_ij_av (off-diagonal) and S_ii=S_i (diagonal). Raises error if minus infinity.
# assumes C_maxsize > 2
def cost_function_simple(partition, correlations_table, alpha, C_maxsize):
    no_of_clusters = len(partition)
    S = np.zeros((no_of_clusters, no_of_clusters))
    val = 0
    for C_i in partition:
        try:
            val = val - alpha * f_clust_square_sharp(len(C_i), C_maxsize)
        except:
            raise InfinityException

    for i in range(no_of_clusters):
        C_i = partition[i]
        S_i = get_S_i(C_i, correlations_table)
        S[i, i] = S_i
        val = val + S_i

    for i in range(no_of_clusters - 1):
        for j in range(i + 1, no_of_clusters):
            C_i = partition[i]
            C_j = partition[j]
            S_ij = get_S_ij_av(C_i, C_j, correlations_table)
            S[i, j] = S_ij
            S[j, i] = S_ij
            val = val - S_ij
    return val, S


def evaluate_move_operation_naive_cummulative(partition, index_k, index_C_i, index_C_j, correlations_table, alpha,
                                              C_maxsize):
    partition_copy = copy.deepcopy(partition)
    val1 = cost_function_simple_cummulative(partition_copy, correlations_table, alpha, C_maxsize)
    make_move_operation(partition_copy, index_k, index_C_i, index_C_j)
    # print(S_1)
    try:
        val2 = cost_function_simple_cummulative(partition_copy, correlations_table, alpha, C_maxsize)
    except InfinityException:
        return 0, False
    diff = val2 - val1
    return diff, True


def evaluate_move_operation_naive(partition, index_k, index_C_i, index_C_j, correlations_table, alpha, C_maxsize, S):
    partition_copy = copy.deepcopy(partition)
    val1, S_1 = cost_function_simple(partition_copy, correlations_table, alpha, C_maxsize)
    # print(S_1)
    if len(partition_copy[index_C_i]) < 2:
        print("deletion")
        S_1 = np.delete(S_1, index_C_i, 0)
        S_1 = np.delete(S_1, index_C_i, 1)

    make_move_operation(partition_copy, index_k, index_C_i, index_C_j)
    # print(S_1)
    try:
        val2, S_2 = cost_function_simple(partition_copy, correlations_table, alpha, C_maxsize)
    except InfinityException:
        return 0, False
    diff = val2 - val1
    dS = S_2 - S_1
    print(S_2)
    return diff, dS, True


def evaluate_swap_operation_naive_cummulative(partition, index_k, index_l, index_C_i, index_C_j, correlations_table,
                                              alpha, C_maxsize):
    partition_copy = copy.deepcopy(partition)
    val1 = cost_function_simple_cummulative(partition_copy, correlations_table, alpha, C_maxsize)
    make_swap_operation(partition_copy, index_k, index_l, index_C_i, index_C_j)
    val2 = cost_function_simple_cummulative(partition_copy, correlations_table, alpha, C_maxsize)
    diff = val2 - val1
    return diff


def evaluate_swap_operation_naive(partition, index_k, index_l, index_C_i, index_C_j, correlations_table, alpha,
                                  C_maxsize, S):
    partition_copy = copy.deepcopy(partition)
    val1, S_1 = cost_function_simple(partition_copy, correlations_table, alpha, C_maxsize)
    make_swap_operation(partition_copy, index_k, index_l, index_C_i, index_C_j)
    val2, S_2 = cost_function_simple(partition_copy, correlations_table, alpha, C_maxsize)
    diff = val2 - val1
    dS = S_2 - S_1
    return diff, dS


def evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j, correlations_table, alpha, C_maxsize):
    C_i = partition[index_C_i]
    C_j = partition[index_C_j]
    c_i = len(C_i)
    c_j = len(C_j)

    diff = 0
    # cluster function part
    dS_clust = 0
    try:
        dS_clust = f_clust_square_sharp(c_i, C_maxsize) + f_clust_square_sharp(c_j, C_maxsize) - f_clust_square_sharp(
            c_i - 1, C_maxsize) - f_clust_square_sharp(c_j + 1, C_maxsize)
        dS_clust = alpha * dS_clust
    except InfinityException:
        return diff, False

    # intra cluster part
    dS_intra = 0
    for index_q in C_j:
        dS_intra = dS_intra + correlations_table[index_q][index_k] + correlations_table[index_k][index_q]
    for index_q in C_i:
        if (index_q == index_k):
            continue
        dS_intra = dS_intra - correlations_table[index_q][index_k] - correlations_table[index_k][index_q]

    diff = dS_clust + dS_intra
    return diff, True


# returns the change of cost function cost_function_simple after a move operation (qubit k from C_i to C_j) on clusters.
def evaluate_move_operation(partition, index_k, index_C_i, index_C_j, correlations_table, alpha, C_maxsize, S):
    no_of_clusters = len(partition)
    dS = np.zeros((no_of_clusters, no_of_clusters))
    diff = 0
    C_i = partition[index_C_i]
    C_j = partition[index_C_j]
    c_i = len(C_i)
    c_j = len(C_j)

    # intra attraction part
    dS_intra = 0

    if c_i > 2:
        psum = 0
        for index_q in C_i:
            if index_q == index_k:
                continue
            psum = psum + correlations_table[index_k][index_q] + correlations_table[index_q][index_k]
        psum = -psum / ((c_i - 1) * (c_i - 2))
        psum = psum + (c_i * (c_i - 1) / ((c_i - 1) * (c_i - 2)) - 1) * S[index_C_i, index_C_i]
    else:
        psum = -S[index_C_i, index_C_i]

    dS[index_C_i, index_C_i] = psum
    dS_intra = dS_intra + psum

    psum = 0
    sum_k_q_C_j = 0
    for index_q in C_j:
        psum = psum + correlations_table[index_k][index_q] + correlations_table[index_q][index_k]
    sum_k_q_C_j = psum
    psum = psum / ((c_j + 1) * c_j)
    psum = psum + (c_j * (c_j - 1) / ((c_j + 1) * c_j) - 1) * S[index_C_j, index_C_j]
    dS[index_C_j, index_C_j] = psum
    dS_intra = dS_intra + psum

    # inter attraction part
    dS_inter = 0
    for index_C_m in range(len(partition)):
        if index_C_m in [index_C_i, index_C_j]:
            continue
        C_m = partition[index_C_m]
        c_m = len(C_m)

        psum = 0
        for index_q in C_m:
            psum = psum + correlations_table[index_k][index_q] + correlations_table[index_q][index_k]
        if c_i > 1:
            psum1 = (-1 / (2 * c_m)) * (1 / (c_i - 1)) * psum + (c_i * c_m / ((c_i - 1) * c_m) - 1) * S[
                index_C_i, index_C_m]
        else:
            psum1 = -S[index_C_i, index_C_m]
        dS[index_C_i, index_C_m] = psum1
        dS[index_C_m, index_C_i] = psum1
        psum2 = (1 / (2 * c_m)) * (1 / (c_j + 1)) * psum + (c_j * c_m / ((c_j + 1) * c_m) - 1) * S[index_C_j, index_C_m]
        dS[index_C_j, index_C_m] = psum2
        dS[index_C_m, index_C_j] = psum2
        dS_inter = dS_inter + psum1 + psum2

    psum = 0
    if c_i > 1:
        for index_q in C_i:
            if index_q == index_k:
                continue
            psum = psum + correlations_table[index_k][index_q] + correlations_table[index_q][index_k]
        psum = (1 / (2 * (c_i - 1) * (c_j + 1))) * (psum - sum_k_q_C_j) + (c_i * c_j / ((c_i - 1) * (c_j + 1)) - 1) * S[
            index_C_i, index_C_j]
    else:
        psum = -S[index_C_i, index_C_j]

    dS[index_C_i, index_C_j] = psum
    dS[index_C_j, index_C_i] = psum
    dS_inter = dS_inter + psum

    # cluster function part
    dS_clust = 0
    try:
        dS_clust = f_clust_square_sharp(len(C_i) - 1, C_maxsize) + f_clust_square_sharp(len(C_j) + 1,
                                                                                        C_maxsize) - f_clust_square_sharp(
            len(C_i),
            C_maxsize) - f_clust_square_sharp(
            len(C_j), C_maxsize)
        dS_clust = alpha * dS_clust
    except InfinityException:
        return diff, dS, False

    diff = dS_intra - dS_inter - dS_clust

    return diff, dS, True


def evaluate_swap_operation_cummulative(partition, index_k, index_l, index_C_i, index_C_j, correlations_table, alpha):
    C_i = partition[index_C_i]
    C_j = partition[index_C_j]

    diff = 0
    # intra cluster part
    dS_intra = 0
    for index_q in C_i:
        if (index_q == index_k):
            continue
        dS_intra = dS_intra + correlations_table[index_q][index_l] + correlations_table[index_l][index_q] - \
                   correlations_table[index_q][index_k] - correlations_table[index_k][index_q]
    for index_q in C_j:
        if (index_q == index_l):
            continue
        dS_intra = dS_intra + correlations_table[index_q][index_k] + correlations_table[index_k][index_q] - \
                   correlations_table[index_q][index_l] - correlations_table[index_l][index_q]
        diff = dS_intra
    return diff


# returns the change of cost function cost_function_simple after a swap operation (qubit k from C_i wilh qubit l from C_j) on clusters.
def evaluate_swap_operation(partition, index_k, index_l, index_C_i, index_C_j, correlations_table, alpha, C_maxsize, S):
    # cluster function part - contrubutes zero
    # intra attraction part
    no_of_clusters = len(partition)
    dS = np.zeros((no_of_clusters, no_of_clusters))
    dS_intra = 0
    C_i = partition[index_C_i]
    C_j = partition[index_C_j]
    c_i = len(C_i)
    c_j = len(C_j)

    if c_i > 1:
        psum = 0
        for index_q in C_i:
            if index_q == index_k:
                continue
            psum = psum + correlations_table[index_q][index_l] + correlations_table[index_l][index_q] - \
                   correlations_table[index_q][index_k] - correlations_table[index_k][index_q]
        psum = psum * (1 / (c_i * (c_i - 1)))
        dS[index_C_i, index_C_i] = psum
        dS_intra = dS_intra + psum

    if c_j > 1:
        psum = 0
        for index_q in C_j:
            if index_q == index_l:
                continue
            psum = psum + correlations_table[index_q][index_k] + correlations_table[index_k][index_q] - \
                   correlations_table[index_q][index_l] - correlations_table[index_l][index_q]
        dS[index_C_j][index_C_j] = psum
        psum = psum * (1 / (c_j * (c_j - 1)))
        dS_intra = dS_intra + psum

    # inter attraction part
    dS_inter = 0
    for index_C_m in range(len(partition)):
        if index_C_m in [index_C_i, index_C_j]:
            continue
        C_m = partition[index_C_m]
        c_m = len(C_m)

        psum = 0
        for index_q in C_m:
            psum = psum + correlations_table[index_l][index_q] + correlations_table[index_q][index_l] - \
                   correlations_table[index_k][index_q] - correlations_table[index_q][index_k]
        psum1 = (1 / (2 * c_m)) * (1 / c_i) * psum
        psum2 = (1 / (2 * c_m)) * (-1 / c_j) * psum
        dS[index_C_i, index_C_m] = psum1
        dS[index_C_m, index_C_i] = psum1
        dS[index_C_j, index_C_m] = psum2
        dS[index_C_m, index_C_j] = psum2
        dS_inter = dS_inter + psum1 + psum2

    psum = 0
    for index_q in C_j:
        if index_q == index_l:
            continue
        psum = psum + correlations_table[index_l][index_q] + correlations_table[index_q][index_l] - \
               correlations_table[index_k][index_q] - correlations_table[index_q][index_k]
    for index_q in C_i:
        if index_q == index_k:
            continue
        psum = psum - (correlations_table[index_l][index_q] + correlations_table[index_q][index_l] -
                       correlations_table[index_k][index_q] - correlations_table[index_q][index_k])
    psum = psum * (1 / (2 * c_i * c_j))
    dS[index_C_i, index_C_j] = psum
    dS[index_C_j, index_C_i] = psum
    dS_inter = dS_inter + psum

    diff = dS_intra - dS_inter
    return diff, dS


def get_initial_partition(correlations_table):
    n = correlations_table.shape[0]
    CL = dict()
    for i in range(n - 1):
        for j in range(i + 1, n):
            val = 0
            val1 = correlations_table[i, j]
            val2 = correlations_table[j, i]
            if (val1 > val2):
                val = val1
            else:
                val = val2
            CL.update({(i, j): val})
    # print(CL)
    partition = []
    qubits = list(range(n))
    while len(qubits) > 1:
        pair = max(CL, key=lambda key: CL[key])
        i = pair[0]
        j = pair[1]
        # print(i,j)
        partition.append([i, j])
        keys = list(CL.keys())
        # print(keys)
        for pair in keys:
            # print(str(pair)+"keys:"+str(i)+str(j))
            if pair[0] in [i, j] or pair[1] in [i, j]:
                # print("popping"+str(pair))
                CL.pop(pair, None)
        qubits.remove(i)
        qubits.remove(j)
    if len(qubits) == 1:
        partition.append([qubits[0]])

    return partition


def return_cluster_index(partition, target_qubit):
    index = 0
    for cluster in partition:
        for qubit in cluster:
            if qubit == target_qubit:
                return index
        index = index + 1
    return index


def make_move_operation(partition, index_k, index_C_i, index_C_j):
    # partition[index_C_i].remove(index_k)
    partition[index_C_j].append(index_k)

    if len(partition[index_C_i]) == 1:
        partition.pop(index_C_i)
    else:
        partition[index_C_i].remove(index_k)

    return


def make_swap_operation(partition, index_k, index_l, index_C_i, index_C_j):
    partition[index_C_i].remove(index_k)
    partition[index_C_i].append(index_l)
    partition[index_C_j].remove(index_l)
    partition[index_C_j].append(index_k)

    return




def choose_nonpos_op(val_of_ops_nonpos, temp):
    """
    Chooses the cluster index for which to move the qubit among all those resulting in nonpositive change of objective function phi.
    Probability of the choice of cluster i is:
    p(i)=exp(change_i/temp)/(no_of_ops_nonpos + epsilon)
    where no_of_ops_nonpos is the number of all operations and epsilon is small number added for numerical stability (here set as 0.001).
    Additionally, no operation can be selected with probability:
    p(reject)= 1- sum_j p(j).

    :param val_of_ops_nonpos: dictionary with keys being cluster indices and values being the changes of phi.
    :type val_of_ops_nonpos: dict

    :param temp: temperature parameter
    :type temp: float

    :return: chosen cluster index.
    :rtype: int
    """

    prob_dict = dict()
    no_of_ops_nonpos = len(val_of_ops_nonpos.keys())
    psum = 0
    for op in val_of_ops_nonpos:
        p = np.exp(val_of_ops_nonpos[op] / temp) / (no_of_ops_nonpos + 0.001)
        prob_dict[op] = p
        psum += p
    prob_dict['reject'] = 1.0 - psum
    chosen_op = random.choices(list(prob_dict.keys()), weights=prob_dict.values(), k=1)[0]
    if chosen_op == 'reject':
        raise RejectionException
    return chosen_op


# requires C_maxsize > 2 (otherwise terminates before execution)
# uses f_clust_square_sharp, cost_function_simple, and both naive non-naive move and swap evaluation (non-cummulative) operations (uses non-naive, naive used only for comparison)
def partition_algorithm_v1(correlations_table, alpha, C_maxsize, N_alg, printing, drawing):
    if (C_maxsize < 3):
        print("Error: max cluster size has to be at least 3!. Algorithm terminated.")
        return
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf, init_S = cost_function_simple(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    # results=dict()
    global_best_parition = initial_partition
    global_best_value = init_cf
    for attempt_no in range(N_alg):
        if (printing):
            print("attempt: " + str(attempt_no + 1))
        partition = copy.deepcopy(initial_partition)
        best_cf = init_cf - 1
        curr_cf = init_cf
        curr_S = init_S
        epoch_no = 0
        while curr_cf > best_cf:
            epoch_no = epoch_no + 1
            best_cf = curr_cf
            if (printing):
                print("starting epoch: " + str(epoch_no))
            pairs = []
            for i in range(no_of_qubits - 1):
                for j in range(i + 1, no_of_qubits):
                    pairs.append([i, j])
            while len(pairs) > 0:
                val_of_ops = dict()
                pair = random.choice(pairs)
                pairs.remove(pair)
                i = pair[0]
                j = pair[1]
                if (printing):
                    print("pair: " + "(" + str(i) + ", " + str(j) + ")")
                index_C_i = return_cluster_index(partition, i)
                index_C_j = return_cluster_index(partition, j)
                if (index_C_i == index_C_j):
                    if (printing):
                        print("WRONG PAIR - SKIP")
                    continue
                index_k = i
                val1, dS1, not_neg_infty = evaluate_move_operation(partition, index_k, index_C_i, index_C_j,
                                                                   correlations_table, alpha, C_maxsize, curr_S)
                # val1_true, dS1_true, not_neg_infty_true = evaluate_move_operation_naive(partition, index_k, index_C_i,index_C_j, correlations_table, alpha, C_maxsize, curr_S)
                if (printing):
                    print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                        index_C_j) + "\t, result: " + "{:.5f}".format(val1) + "\t, size_leq_C_max: " + str(
                        not_neg_infty))
                if val1 > 0 and not_neg_infty:
                    if (printing):
                        print("update")
                    val_of_ops.update({'move_ij': val1})
                index_C_i = return_cluster_index(partition, j)
                index_C_j = return_cluster_index(partition, i)
                index_k = j
                val2, dS2, not_neg_infty = evaluate_move_operation(partition, index_k, index_C_i, index_C_j,
                                                                   correlations_table, alpha, C_maxsize, curr_S)
                # val2_true, dS2_true, not_neg_infty_true = evaluate_move_operation_naive(partition, index_k, index_C_i, index_C_j, correlations_table, alpha, C_maxsize, curr_S)
                if (printing):
                    print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                        index_C_j) + "\t, result: " + "{:.5f}".format(val2) + "\t, size_leq_C_max: " + str(
                        not_neg_infty))
                if val2 > 0 and not_neg_infty:
                    val_of_ops.update({'move_ji': val2})
                index_C_i = return_cluster_index(partition, i)
                index_C_j = return_cluster_index(partition, j)
                index_k = i
                index_l = j
                val3, dS3 = evaluate_swap_operation(partition, index_k, index_l, index_C_i, index_C_j,
                                                    correlations_table, alpha, C_maxsize, curr_S)
                val3_true, dS3_true = evaluate_swap_operation_naive(partition, index_k, index_l, index_C_i, index_C_j,
                                                                    correlations_table, alpha, C_maxsize, curr_S)
                if (printing):
                    print("checking swap " + str(index_C_i) + "<-" + str(index_k) + "&" + str(index_l) + "->" + str(
                        index_C_j) + "\t, result: " + "{:.5f}".format(val3) + "\t, size_leq_C_max: " + str(
                        not_neg_infty) + "\t, true_res:" + "{:.5f}".format(val3_true))
                if val3 > 0:
                    val_of_ops.update({'swap': val3})

                if len(val_of_ops) > 0:
                    if (printing):
                        print("ACCEPT:")
                    op = max(val_of_ops, key=lambda key: val_of_ops[key])
                    if (op == 'move_ij'):
                        dS = dS1
                        index_C_i = return_cluster_index(partition, i)
                        index_C_j = return_cluster_index(partition, j)
                        index_k = i
                        c_i = len(partition[index_C_i])
                        if (printing):
                            print("MOVE_ij")
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        print(partition)
                        curr_cf = curr_cf + val_of_ops[op]
                        curr_S = curr_S + dS
                        if c_i < 2:
                            curr_S = np.delete(curr_S, index_C_i, 0)
                            curr_S = np.delete(curr_S, index_C_i, 1)
                    elif (op == 'move_ji'):
                        dS = dS2
                        index_C_i = return_cluster_index(partition, j)
                        index_C_j = return_cluster_index(partition, i)
                        index_k = j
                        c_i = len(partition[index_C_i])
                        if (printing):
                            print("MOVE_ji")
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        print(partition)
                        curr_cf = curr_cf + val_of_ops[op]
                        curr_S = curr_S + dS
                        if c_i < 2:
                            curr_S = np.delete(curr_S, index_C_i, 0)
                            curr_S = np.delete(curr_S, index_C_i, 1)
                    else:
                        dS = dS3
                        index_C_i = return_cluster_index(partition, i)
                        index_C_j = return_cluster_index(partition, j)
                        index_k = i
                        index_l = j
                        if (printing):
                            print("SWAP")
                        make_swap_operation(partition, index_k, index_l, index_C_i, index_C_j)
                        print(partition)
                        curr_cf = curr_cf + val_of_ops[op]
                        curr_S = curr_S + dS
                else:
                    if (printing):
                        print("REJECT ALL")
        if (printing):
            print("convergence to: " + str(best_cf) + " after " + str(epoch_no) + " epochs")
        if (best_cf > global_best_value):
            global_best_parition = partition
            global_best_value = best_cf
        # results.update({partition:best_cf})
    # best_attempt=op=max(results, key=lambda key: results[key])
    return global_best_parition, global_best_value


# requires C_maxsize >= 2 (otherwise terminates before execution)
# uses f_clust_square_sharp, cost_function_simple_cummulative, and non-naive move and swap cummulative evaluation operations
def partition_algorithm_v1_cummulative(correlations_table,
                                       alpha,
                                       C_maxsize,
                                       N_alg,
                                       printing,
                                       drawing, disable_pb):
    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        return
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = cost_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    # results=dict()
    global_best_parition = initial_partition
    global_best_value = init_cf
    if (skip_optimisation == False):
        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            best_cf = init_cf - 1
            curr_cf = init_cf
            epoch_no = 0
            while curr_cf > best_cf:
                epoch_no = epoch_no + 1
                best_cf = curr_cf
                if (printing):
                    print("starting epoch: " + str(epoch_no))
                pairs = []
                for i in range(no_of_qubits - 1):
                    for j in range(i + 1, no_of_qubits):
                        pairs.append([i, j])
                while len(pairs) > 0:
                    val_of_ops = dict()
                    pair = random.choice(pairs)
                    pairs.remove(pair)
                    i = pair[0]
                    j = pair[1]
                    if (printing):
                        print("pair: " + "(" + str(i) + ", " + str(j) + ")")
                    index_C_i = return_cluster_index(partition, i)
                    index_C_j = return_cluster_index(partition, j)
                    if (index_C_i == index_C_j):
                        if (printing):
                            print("WRONG PAIR - SKIP")
                        continue
                    index_k = i
                    val1, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                              correlations_table, alpha, C_maxsize)
                    # val1_true, not_neg_infty_true = evaluate_move_operation_naive_cummulative(partition, index_k, index_C_i, index_C_j, correlations_table, alpha, C_maxsize)

                    if (printing):
                        print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                            index_C_j) + "\t, result: " + "{:.5f}".format(val1) + "\t, size_leq_C_max: " + str(
                            not_neg_infty))
                    if val1 > 0 and not_neg_infty:
                        val_of_ops.update({'move_ij': val1})
                    index_C_i = return_cluster_index(partition, j)
                    index_C_j = return_cluster_index(partition, i)
                    index_k = j
                    val2, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                              correlations_table, alpha, C_maxsize)
                    # val2_true, not_neg_infty_true = evaluate_move_operation_naive_cummulative(partition, index_k, index_C_i, index_C_j, correlations_table, alpha, C_maxsize)
                    if (printing):
                        print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                            index_C_j) + "\t, result: " + "{:.5f}".format(val2) + "\t, size_leq_C_max: " + str(
                            not_neg_infty))
                    if val2 > 0 and not_neg_infty:
                        val_of_ops.update({'move_ji': val2})
                    index_C_i = return_cluster_index(partition, i)
                    index_C_j = return_cluster_index(partition, j)
                    index_k = i
                    index_l = j
                    val3 = evaluate_swap_operation_cummulative(partition, index_k, index_l, index_C_i, index_C_j,
                                                               correlations_table, alpha)
                    # val3_true = evaluate_swap_operation_naive_cummulative(partition, index_k, index_l, index_C_i, index_C_j, correlations_table, alpha, C_maxsize)
                    if (printing):
                        print("checking swap " + str(index_C_i) + "<-" + str(index_k) + "&" + str(index_l) + "->" + str(
                            index_C_j) + "\t, result: " + "{:.5f}".format(val3) + "\t, size_leq_C_max: " + str(
                            not_neg_infty))
                    if val3 > 0:
                        val_of_ops.update({'swap': val3})

                    if len(val_of_ops) > 0:
                        if (printing):
                            print("ACCEPT:")
                        op = max(val_of_ops, key=lambda key: val_of_ops[key])
                        if (op == 'move_ij'):
                            index_C_i = return_cluster_index(partition, i)
                            index_C_j = return_cluster_index(partition, j)
                            index_k = i
                            make_move_operation(partition, index_k, index_C_i, index_C_j)
                            if (printing):
                                print("MOVE_ij")
                                print(partition)
                            curr_cf = curr_cf + val_of_ops[op]
                        elif (op == 'move_ji'):
                            index_C_i = return_cluster_index(partition, j)
                            index_C_j = return_cluster_index(partition, i)
                            index_k = j
                            make_move_operation(partition, index_k, index_C_i, index_C_j)
                            if (printing):
                                print("MOVE_ji")
                                print(partition)
                            curr_cf = curr_cf + val_of_ops[op]
                        else:
                            index_C_i = return_cluster_index(partition, i)
                            index_C_j = return_cluster_index(partition, j)
                            index_k = i
                            index_l = j
                            make_swap_operation(partition, index_k, index_l, index_C_i, index_C_j)
                            if (printing):
                                print("SWAP")
                                print(partition)
                            curr_cf = curr_cf + val_of_ops[op]
                    elif (printing):
                        print("REJECT ALL")
            if (printing):
                print("convergence to: " + str(best_cf) + " after " + str(epoch_no) + " epochs")
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value


# requires C_maxsize >= 2 (otherwise terminates before execution)
# uses f_clust_square_sharp, cost_function_simple_cummulative, and non-naive move and swap cummulative evaluation operations
# TODO OS: add optional convergence detection (termination after N_rand steps or if convergence detected). For example if a pair is rejected it goes to a list. If pair is accepted the list is erased. If list is full the algorithm terminates current attempt.
def partition_algorithm_v2_cummulative(correlations_table,
                                       alpha,
                                       C_maxsize,
                                       N_alg,
                                       N_rand,
                                       printing,
                                       drawing, disable_pb):
    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        return
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = cost_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    # results=dict()
    global_best_parition = initial_partition
    global_best_value = init_cf

    if (skip_optimisation == False):

        pairs = []
        for i in range(no_of_qubits - 1):
            for j in range(i + 1, no_of_qubits):
                pairs.append([i, j])

        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            best_cf = init_cf - 1
            curr_cf = init_cf
            move_no = 0

            while move_no <= N_rand - 1:
                move_no = move_no + 1
                best_cf = curr_cf
                if (printing):
                    print("starting move: " + str(move_no))

                val_of_ops = dict()
                pair = random.choice(pairs)
                i = pair[0]
                j = pair[1]
                if (printing):
                    print("pair: " + "(" + str(i) + ", " + str(j) + ")")
                index_C_i = return_cluster_index(partition, i)
                index_C_j = return_cluster_index(partition, j)
                if (index_C_i == index_C_j):
                    if (printing):
                        print("WRONG PAIR - SKIP")
                    continue
                index_k = i
                val1, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                          correlations_table, alpha, C_maxsize)
                # val1_true, not_neg_infty_true = evaluate_move_operation_naive_cummulative(partition, index_k, index_C_i, index_C_j, correlations_table, alpha, C_maxsize)

                if (printing):
                    print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                        index_C_j) + "\t, result: " + "{:.5f}".format(val1) + "\t, size_leq_C_max: " + str(
                        not_neg_infty))
                if val1 > 0 and not_neg_infty:
                    val_of_ops.update({'move_ij': val1})
                index_C_i = return_cluster_index(partition, j)
                index_C_j = return_cluster_index(partition, i)
                index_k = j
                val2, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                          correlations_table, alpha, C_maxsize)
                # val2_true, not_neg_infty_true = evaluate_move_operation_naive_cummulative(partition, index_k, index_C_i, index_C_j, correlations_table, alpha, C_maxsize)
                if (printing):
                    print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                        index_C_j) + "\t, result: " + "{:.5f}".format(val2) + "\t, size_leq_C_max: " + str(
                        not_neg_infty))
                if val2 > 0 and not_neg_infty:
                    val_of_ops.update({'move_ji': val2})
                index_C_i = return_cluster_index(partition, i)
                index_C_j = return_cluster_index(partition, j)
                index_k = i
                index_l = j
                val3 = evaluate_swap_operation_cummulative(partition, index_k, index_l, index_C_i, index_C_j,
                                                           correlations_table, alpha)
                # val3_true = evaluate_swap_operation_naive_cummulative(partition, index_k, index_l, index_C_i, index_C_j, correlations_table, alpha, C_maxsize)
                if (printing):
                    print("checking swap " + str(index_C_i) + "<-" + str(index_k) + "&" + str(index_l) + "->" + str(
                        index_C_j) + "\t, result: " + "{:.5f}".format(val3) + "\t, size_leq_C_max: " + str(
                        not_neg_infty))
                if val3 > 0:
                    val_of_ops.update({'swap': val3})

                if len(val_of_ops) > 0:
                    if (printing):
                        print("ACCEPT:")
                    op = max(val_of_ops, key=lambda key: val_of_ops[key])
                    if (op == 'move_ij'):
                        index_C_i = return_cluster_index(partition, i)
                        index_C_j = return_cluster_index(partition, j)
                        index_k = i
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        if (printing):
                            print("MOVE_ij")
                            print(partition)
                        curr_cf = curr_cf + val_of_ops[op]
                    elif (op == 'move_ji'):
                        index_C_i = return_cluster_index(partition, j)
                        index_C_j = return_cluster_index(partition, i)
                        index_k = j
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        if (printing):
                            print("MOVE_ji")
                            print(partition)
                        curr_cf = curr_cf + val_of_ops[op]
                    else:
                        index_C_i = return_cluster_index(partition, i)
                        index_C_j = return_cluster_index(partition, j)
                        index_k = i
                        index_l = j
                        make_swap_operation(partition, index_k, index_l, index_C_i, index_C_j)
                        if (printing):
                            print("SWAP")
                            print(partition)
                        curr_cf = curr_cf + val_of_ops[op]
                elif (printing):
                    print("REJECT ALL")
            if (printing):
                print("Obtained: " + str(best_cf) + " after " + str(move_no) + " moves")
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value


# requires C_maxsize >= 2 (otherwise terminates before execution)
# uses f_clust_square_sharp, cost_function_simple_cummulative, and non-naive move and swap cummulative evaluation operations
# TODO OS: add optional convergence detection (termination after N_rand steps or if convergence detected)
def partition_algorithm_v3_cummulative(correlations_table,
                                       alpha,
                                       C_maxsize,
                                       N_alg,
                                       N_rand,
                                       printing,
                                       drawing, disable_pb):
    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        return
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = cost_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    # results=dict()
    global_best_parition = initial_partition
    global_best_value = init_cf

    if (skip_optimisation == False):

        qubits = range(no_of_qubits)

        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            # best_cf = init_cf - 1
            best_cf = init_cf
            move_no = 0

            while move_no <= N_rand - 1:
                move_no = move_no + 1
                # best_cf = curr_cf
                if (printing):
                    print("starting move: " + str(move_no))

                val_of_ops = dict()
                i = random.choice(qubits)
                if (printing):
                    print("selected qubit: " + str(i))
                index_C_i = return_cluster_index(partition, i)
                index_k = i
                for index_C_j in range(len(partition)):
                    if index_C_j == index_C_i:
                        continue

                    val, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                             correlations_table, alpha, C_maxsize)

                    if (printing):
                        print("checking move of qubit " + str(i) + " to partition " + str(
                            index_C_j) + "\t, result: " + "{:.5f}".format(val) + "\t, size_leq_C_max: " + str(
                            not_neg_infty))
                    if val > 0 and not_neg_infty:
                        val_of_ops.update({index_C_j: val})
                if len(val_of_ops) > 0:
                    best_cluster_index = max(val_of_ops, key=lambda key: val_of_ops[key])
                    best_cf = best_cf + val_of_ops[best_cluster_index]
                    index_C_j = best_cluster_index
                    make_move_operation(partition, index_k, index_C_i, index_C_j)
                    if (printing):
                        print("ACCEPT move to cluster no " + str(best_cluster_index) + ", value=" + str(best_cf))
                        print(partition)
                elif (printing):
                    print("REJECT ALL")

            if (printing):
                print("Obtained: " + str(best_cf) + " after move " + str(move_no))
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value


def partition_algorithm_v4_cummulative(correlations_table,
                                       alpha,
                                       C_maxsize,
                                       N_alg,
                                       printing,
                                       drawing, disable_pb):
    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        return
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = cost_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    # results=dict()
    global_best_parition = initial_partition
    global_best_value = init_cf
    if (skip_optimisation == False):
        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            best_cf = init_cf - 1
            curr_cf = init_cf
            epoch_no = 0
            while curr_cf > best_cf:
                epoch_no = epoch_no + 1
                best_cf = curr_cf
                if (printing):
                    print("starting epoch: " + str(epoch_no))
                qubits = list(range(no_of_qubits))

                while len(qubits) > 0:
                    val_of_ops = dict()
                    i = random.choice(qubits)
                    qubits.remove(i)

                    if (printing):
                        print("selected qubit: " + str(i))
                    index_C_i = return_cluster_index(partition, i)
                    index_k = i
                    for index_C_j in range(len(partition)):
                        if index_C_j == index_C_i:
                            continue

                        val, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i,
                                                                                 index_C_j, correlations_table, alpha,
                                                                                 C_maxsize)

                        if (printing):
                            print("checking move of qubit " + str(i) + " to partition " + str(
                                index_C_j) + "\t, result: " + "{:.5f}".format(val) + "\t, size_leq_C_max: " + str(
                                not_neg_infty))
                        if val > 0 and not_neg_infty:
                            val_of_ops.update({index_C_j: val})
                    if len(val_of_ops) > 0:
                        best_cluster_index = max(val_of_ops, key=lambda key: val_of_ops[key])
                        curr_cf = curr_cf + val_of_ops[best_cluster_index]
                        index_C_j = best_cluster_index
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        if (printing):
                            print("ACCEPT move to cluster no " + str(best_cluster_index) + ", value=" + str(curr_cf))
                            print(partition)
                    elif (printing):
                        print("REJECT ALL")

            if (printing):
                print("convergence to: " + str(best_cf) + " after " + str(epoch_no) + " epochs")
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value



# uses f_clust_square_sharp, objective_function_simple_cummulative, and non-naive move and swap cummulative evaluation operations
def partition_algorithm_v4_cummulative_temp(correlations_table,
                                            alpha,
                                            C_maxsize,
                                            N_alg,
                                            printing,
                                            drawing, disable_pb, temp=1.0):
    """
    Partition algorithm using move operations on qubits with termination criterion and metropolis-like acceptance condition (regulated by temp).
    If C_maxsize < 2 then algorithm terminates with exception raised. If C_maxsize = 2 then the algorithm is not executed and the initial partition is returned.
    The algorith is run N_alg times and the best partition is chosen. In each run, the algorithm traverses qubits in epochs.
    An epoch consists of all qubits that are traversed at random and without repetitions.
    The partition is being updated during each epoch, by executing move operations on qubits.
    The algorithm terminates, if after some epoch the value of objective function phi has not improved compared to the end of a previous epoch.

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :param N_alg: Number of runs.
    :type N_alg: int

    :param printing: Is printing allowed.
    :type printing: bool

    :param drawing: Is cluster drawing allowed.
    :type drawing: bool

    :param disable_pb: Is progress bar disabled.
    :type disable_pb: bool

    :param temp: Temperture parameter (default is 1.0).
    :type temp: float

    :return best partition found, value of the objective function for this partition
    :rtype:(List[int], float)

    :raises MaxClusterSizeException: returned if max allowed cluster size is < 2
    """

    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        raise MaxClusterSizeException
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = cost_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    global_best_parition = initial_partition
    global_best_value = init_cf
    if (skip_optimisation == False):
        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            best_cf = init_cf - 1
            curr_cf = init_cf
            epoch_no = 0
            while curr_cf > best_cf:
                epoch_no = epoch_no + 1
                best_cf = curr_cf
                if (printing):
                    print("starting epoch: " + str(epoch_no))
                qubits = list(range(no_of_qubits))

                while len(qubits) > 0:
                    val_of_ops_pos = dict()
                    val_of_ops_nonpos = dict()
                    i = random.choice(qubits)
                    qubits.remove(i)

                    if (printing):
                        print("selected qubit: " + str(i))
                    index_C_i = return_cluster_index(partition, i)
                    index_k = i
                    for index_C_j in range(len(partition)):
                        if index_C_j == index_C_i:
                            continue

                        val, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i,
                                                                                 index_C_j, correlations_table, alpha,
                                                                                 C_maxsize)

                        if (printing):
                            print("checking move of qubit " + str(i) + " to partition " + str(
                                index_C_j) + "\t, result: " + "{:.5f}".format(val) + "\t, size_leq_C_max: " + str(
                                not_neg_infty))
                        if val > 0 and not_neg_infty:
                            val_of_ops_pos.update({index_C_j: val})
                        elif val <= 0 and not_neg_infty:
                            val_of_ops_nonpos.update({index_C_j: val})
                    if len(val_of_ops_pos) > 0:
                        chosen_cluster_index = choose_pos_op(val_of_ops_pos, temp)
                        curr_cf = curr_cf + val_of_ops_pos[chosen_cluster_index]
                        index_C_j = chosen_cluster_index
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        if (printing):
                            print("ACCEPT move to cluster no " + str(chosen_cluster_index) + ", value=" + str(curr_cf))
                            print(partition)
                    else:
                        try:
                            chosen_cluster_index = choose_nonpos_op(val_of_ops_nonpos, temp)
                            curr_cf = curr_cf + val_of_ops_nonpos[chosen_cluster_index]
                            index_C_j = chosen_cluster_index
                            make_move_operation(partition, index_k, index_C_i, index_C_j)
                            if (printing):
                                print("ACCEPT move to cluster no " + str(chosen_cluster_index) + ", value=" + str(
                                    curr_cf))
                                print(partition)
                        except RejectionException:
                            if (printing):
                                print("REJECT ALL")

            if (printing):
                print("convergence to: " + str(best_cf) + " after " + str(epoch_no) + " epochs")
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value

def choose_pos_op(val_of_ops_pos, temp):
    """
    Chooses the cluster index for which to move the qubit among all those resulting in positive change of objective function phi.
    Probability of the choice of cluster i is:
    p(i)=exp((change_i-max_change)/temp)/norm_const
    where norm_const is the normalization constant: sum_j exp((change_j-max_change)/temp.

    :param val_of_ops_pos: dictionary with keys being cluster indices and values being the changes of phi.
    :type val_of_ops_pos: dict

    :param temp: temperature parameter
    :type temp: float

    :return: chosen cluster index.
    :rtype: int
    """

    prob_dict = dict()
    values = np.array(list(val_of_ops_pos.values()))
    max_value = np.max(values)
    exp_args = (values - max_value) / temp
    exp_vals = np.exp(exp_args)
    norm_const = np.sum(exp_vals)
    for op in val_of_ops_pos:
        prob_dict[op] = np.exp((val_of_ops_pos[op] - max_value) / temp) / norm_const
    chosen_op = random.choices(list(prob_dict.keys()), weights=prob_dict.values(), k=1)[0]
    return chosen_op


# -*- coding: utf-8 -*-
"""
The module contains helper functions to work with MOSEK.

Created on Tue Nov 18 10:51:14 2014

@author: Peter Wittek
"""
from __future__ import division, print_function
import numpy as np
import sys
from .sdpa_utils import convert_row_to_sdpa_index


def streamprinter(text):
    """Helper function for printing MOSEK messages in the Python console.
    """
    sys.stdout.write(text)
    sys.stdout.flush()


def moseksol_to_xmat(vec, block_struct):
    n = int(np.sqrt(1+8*len(vec))-1)/2
    if n*(n+1)/2 != len(vec):
        raise ValueError('vec should be of dimension n(n+1)/2')
    M = np.zeros((block_struct[0], block_struct[0]))
    row, column, block, index = 0, 0, 0, 0
    result = []
    while True:
        if column == block_struct[block]:
            row += 1
            column = row
            index += sum(block_struct[block+1:])
        if row == block_struct[block]:
            result.append(M.copy())
            block += 1
            if block == len(block_struct):
                break
            M = np.zeros((block_struct[block], block_struct[block]))
            row, column = 0, 0
        M[row, column] = vec[index]
        M[column, row] = vec[index]
        index += 1
        column += 1
    return result


def parse_mosek_solution(sdp, task):
    import mosek
    soltype = mosek.soltype.itr
    primal, dual = task.getprimalobj(soltype), task.getdualobj(soltype)
    x_mat, y_mat = [], []
    size_ = sum(bs for bs in sdp.block_struct)
    primal_solution = np.zeros(size_*(size_+1) // 2)
    task.getbarxj(soltype, 0, primal_solution)
    y_mat = moseksol_to_xmat(primal_solution, sdp.block_struct)
    dual_solution = np.zeros(size_*(size_+1) // 2)
    task.getbarsj(soltype, 0, dual_solution)
    x_mat = moseksol_to_xmat(dual_solution, sdp.block_struct)
    status = repr(task.getsolsta(soltype))
    return primal, dual, x_mat, y_mat, status


def solve_with_mosek(sdp, solverparameters=None):
    task = convert_to_mosek(sdp)
    if solverparameters is not None:
        import mosek
        for par, val in solverparameters.items():
            try:
                #get rid of a leading "mosek."
                if(par.startswith("mosek.")):
                    par = par[6:]
                #if ?param. prefix is given use it to determine the type
                if(par.startswith("iparam.")):
                    mskpar = eval('mosek.' + par)
                    task.putintparam(mskpar, val)
                elif(par.startswith("dparam.")):
                    mskpar = eval('mosek.' + par)
                    task.putdouparam(mskpar, val)
                elif(par.startswith("sparam.")):
                    mskpar = eval('mosek.' + par)
                    task.putstrparam(mskpar, val)
                #else try to infer the type from the type of val
                elif isinstance(val, int):
                    mskpar = eval('mosek.iparam.' + par)
                    task.putintparam(mskpar, val)
                elif isinstance(val, float):
                    mskpar = eval('mosek.dparam.' + par)
                    task.putdouparam(mskpar, val)
                elif isinstance(val, str):
                    mskpar = eval('mosek.sparam.' + par)
                    task.putstrparam(mskpar, val)
            except AttributeError:
                raise Exception('No mosek parameter found with the name mosek(.iparam|.dparam|.sparam|).'+str(par)+' that takes a value of type '+str(type(val))+'. See the mosek python API manual for valid parameters.')
    task.optimize()
    primal, dual, x_mat, y_mat, status = parse_mosek_solution(sdp,
                                                              task)
    return -primal+sdp.constant_term, \
           -dual+sdp.constant_term, x_mat, y_mat, status


def convert_to_mosek_index(block_struct, row_offsets, block_offsets, row):
    """MOSEK requires a specific sparse format to define the lower-triangular
    part of a symmetric matrix. This function does the conversion from the
    sparse upper triangular matrix format of Ncpol2SDPA.
    """
    block_index, i, j = convert_row_to_sdpa_index(block_struct, row_offsets,
                                                  row)

    offset = block_offsets[block_index]
    ci = offset + i
    cj = offset + j
    return cj, ci  # Note that MOSEK expects lower-triangular matrices


def convert_to_mosek_matrix(sdp):
    """Converts the entire sparse representation of the Fi constraint matrices
    to sparse MOSEK matrices.
    """
    barci = []
    barcj = []
    barcval = []
    barai = []
    baraj = []
    baraval = []
    for k in range(sdp.n_vars):
        barai.append([])
        baraj.append([])
        baraval.append([])
    row_offsets = [0]
    block_offsets = [0]
    cumulative_sum = 0
    cumulative_square_sum = 0
    for block_size in sdp.block_struct:
        cumulative_sum += block_size
        cumulative_square_sum += block_size ** 2
        row_offsets.append(cumulative_square_sum)
        block_offsets.append(cumulative_sum)
    for row in range(len(sdp.F.rows)):
        if len(sdp.F.rows[row]) > 0:
            col_index = 0
            for k in sdp.F.rows[row]:
                value = sdp.F.data[row][col_index]
                i, j = convert_to_mosek_index(sdp.block_struct,
                                              row_offsets, block_offsets, row)
                if k > 0:
                    barai[k - 1].append(i)
                    baraj[k - 1].append(j)
                    baraval[k - 1].append(-value)
                else:
                    barci.append(i)
                    barcj.append(j)
                    barcval.append(value)
                col_index += 1
    return barci, barcj, barcval, barai, baraj, baraval


def convert_to_mosek(sdp):
    """Convert an SDP relaxation to a MOSEK task.

    :param sdp: The SDP relaxation to convert.
    :type sdp: :class:`ncpol2sdpa.sdp`.

    :returns: :class:`mosek.Task`.
    """
    import mosek
    # Cheat when variables are complex and convert with PICOS
    if sdp.complex_matrix:
        from .picos_utils import convert_to_picos
        Problem = convert_to_picos(sdp).to_real()
        Problem._make_mosek_instance()
        task = Problem.msk_task
        if sdp.verbose > 0:
            task.set_Stream(mosek.streamtype.log, streamprinter)
        return task

    barci, barcj, barcval, barai, baraj, baraval = \
        convert_to_mosek_matrix(sdp)
    bkc = [mosek.boundkey.fx] * sdp.n_vars
    blc = [-v for v in sdp.obj_facvar]
    buc = [-v for v in sdp.obj_facvar]

    env = mosek.Env()
    task = env.Task(0, 0)
    if sdp.verbose > 0:
        task.set_Stream(mosek.streamtype.log, streamprinter)
    numvar = 0
    numcon = len(bkc)
    BARVARDIM = [sum(sdp.block_struct)]

    task.appendvars(numvar)
    task.appendcons(numcon)
    task.appendbarvars(BARVARDIM)
    for i in range(numcon):
        task.putconbound(i, bkc[i], blc[i], buc[i])

    symc = task.appendsparsesymmat(BARVARDIM[0], barci, barcj, barcval)
    task.putbarcj(0, [symc], [1.0])

    for i in range(len(barai)):
        syma = task.appendsparsesymmat(BARVARDIM[0], barai[i], baraj[i],
                                       baraval[i])
        task.putbaraij(i, 0, [syma], [1.0])

    # Input the objective sense (minimize/maximize)
    task.putobjsense(mosek.objsense.minimize)

    return task



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 19 15:21:03 2018

@author: flaviobaccari
"""

import numpy as np
from math import fmod
from qrem.ctmp.modeltools.ncpol2sdpa import generate_variables


def generate_commuting_variables(party, label):
    """Generates the list of symbolic variables representing the spins
    for a given site. The variables are treated as commuting.
    :param party: configuration indicating the configuration of number m
                  of measurements and outcomes d for each measurement. It is a
                  list with m integers, each of them representing the number of
                  outcomes of the corresponding  measurement.
    :type party: list of int
    :param label: label to represent the given party
    :type label: str
    :returns: list of sympy.core.symbol.Symbol
    """

    variables = []
    for i in range(len(party)):
        variables.append(
            generate_variables(label + '%s' % i, party[i] - 1, hermitian=True))
    return variables


def get_square2D_spins(k, l, configuration=None):
    """Genrates the list of symbolic variables representing the spins
    arranged in a 2D square lattice of size kxl.
    The variables are treated as commuting.

    :param k: horizontal size of the lattice
    :type k: int
    :param l: vertical size of the lattice
    :type l: int
    :param coniguration: number of spins per site with corresponding levels [d...d]
    :type configuration: list of int

    :returns: list of sympy.core.symbol.Symbol
    """


    if configuration is None:
        configuration = [2]

    s_variables = [[
        generate_commuting_variables(configuration, chr(65 + j * l + i))
        for i in range(l)
    ] for j in range(k)]

    # s_variables = [
    #     generate_commuting_variables(configuration, chr(65 + j * l + i))
    #     for i in range(l)
    # for j in range(k)]

    return s_variables


def get_2Dsquare_localdisorder(k, l, sigma):
    """Generates random local magnetic field values for a 2D square lattice
    distributed according to a Gaussian with zero mean and variance sigma

    :param k: horizontal size of the lattice
    :type k: int
    :param l: vertical size of the lattice
    :type l: int
    :param sigma: variance of the gaussian distribution
    :type sigma: float

    :returns: the magnetic field values h as list of float
    """

    local = []

    for i in range(k):

        row = []
        for j in range(l):

            # Add the local magnetic term

            magnetic = np.random.normal(0, sigma)

            row.append(magnetic)

        local.append(row)

    return local


def get_2Dsquare_ferromagneticdisorder_hamiltonian(k, l, local, s_variables):
    """Generates the hamiltonian of a ferromagnetic 2D Ising model and local
    mangetic fields as a polynomial in the spin variables

    :param k: horizontal size of the lattice
    :type k: int
    :param l: vertical size of the lattice
    :type l: int
    :param local: local magnetic field values
    :type local: list of float
    :param s_variables: list of spin variables
    :type s_variables: list of sympy.core.symbol.Symbol

    :returns: the hamiltonian as sympy.core.add.Add
    """
    hamiltonian = 0
    for i in range(k):

        for j in range(l):

            # Put the link on horizontal line unless I am at l
            if l > 1:
                if j is not (l - 1):

                    hamiltonian += -1.0 * s_variables[i][j][0][
                        0] * s_variables[i][int(fmod(j + 1, l))][0][0]

            # Now the link on the vertical line
            if k > 1:
                if i is not (k - 1):

                    hamiltonian += -1.0 * s_variables[i][j][0][
                        0] * s_variables[int(fmod(i + 1, k))][j][0][0]

            # Add the local magnetic term

            magnetic = local[i][j]
            hamiltonian += magnetic * s_variables[i][j][0][0]

    return hamiltonian


def get_lattice_spins(lattice, configuration):
    """Generates a dictionary of symbolic variables representing the
    spin in the given lattice

    :param lattice: the lattice structure
    :type lattice: networkx.classes.graph.Graph
    :param coniguration: number of spins per site with corresponding levels [d...d]
    :type configuration: list of int

    :returns: the spin variables as dict with tuple as keys representing the nodes
    and sympy.core.symbol.Symbol as values representing the spins
    """
    s_variables = {}
    for i, node in enumerate(lattice.nodes()):

        s_variables[node] = generate_commuting_variables(
            configuration, chr(65 + i))

    return s_variables


def get_lattice_localdisorder(lattice, sigma):
    """Generates the random magneitc field on each node of the lattice, according to
    a gaussian distribution with zero mean and variance sigma

    :param lattice: the lattice structure
    :type lattice: networkx.classes.graph.Graph
    :param sigma: variance of the gaussian distribution
    :type sigma: float

    :returns: the local field as dictionary with tuple as keys representing the nodes
    and float as values
    """

    h = {}

    for node in lattice.nodes():

        h[node] = np.random.normal(0, sigma)

    return h


def get_lattice_ferromangetic(lattice):
    """Generates the ferromagnetic couplings for each edge in the lattice

    :param lattice: the lattice structure
    :type lattice: networkx.classes.graph.Graph

    :returns: the couplings as dictionary with tuple as keys representing the edges
    and float as values
    """

    J = {}

    for edge in lattice.edges():

        J[edge] = -1

    return J


def get_lattice_hamiltonian(lattice, h, J, s_variables):
    """Generates the hamiltonian for the given lattice with the given local fields h
    and couplings J as a polynomial in the spin variables

    :param lattice: the lattice structure
    :type lattice: networkx.classes.graph.Graph
    :param h: the local fields h_i
    :type h: dict
    :param J: the couplings J_{ij}
    :type J: dict
    :param s_variables: list of spin variables
    :type s_variables: list of sympy.core.symbol.Symbol

    :returns: the hamiltonian as sympy.core.add.Add
    """

    hamiltonian = 0

    for node in lattice.nodes():
        hamiltonian += h[node] * s_variables[node][0][0]

    for edge in lattice.edges():
        hamiltonian += J[edge] * s_variables[edge[0]][0][0] * s_variables[edge[
            1]][0][0]

    return hamiltonian



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""

import abc


class EnergyEstimatorInterface(abc.ABC):
    """
    This is interface for classes that will perform error mitigation for energy estimation.
    """

    @property
    @abc.abstractmethod
    def results_dictionary(self) -> dict:
        # dictionary of experimental results
        raise NotImplementedError

    @results_dictionary.setter
    @abc.abstractmethod
    def results_dictionary(self, results_dictionary: dict) -> None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def hamiltonian_dictionaries(self) -> dict:
        # Dictionary storing information about Hamiltonians weights
        raise NotImplementedError

    @hamiltonian_dictionaries.setter
    @abc.abstractmethod
    def hamiltonian_dictionaries(self, hamiltonian_dictionaries: dict) -> None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def energies_dictionary_raw(self) -> dict:
        # Dictionary storing information about estimated energies (raw)
        raise NotImplementedError

    @energies_dictionary_raw.setter
    @abc.abstractmethod
    def energies_dictionary_raw(self,energies_dictionary_raw:dict) -> dict:
        # Dictionary storing information about estimated energies (raw)
        raise NotImplementedError


    # @property
    # @abc.abstractmethod
    # def energies_dictionary_corrected(self) -> dict:
    #     # Dictionary storing information about estimated energies (corrected)
    #     raise NotImplementedError
    #
    # @energies_dictionary_corrected.setter
    # @abc.abstractmethod
    # def energies_dictionary_corrected(self,energies_dictionary_corrected:dict) -> dict:
    #     # Dictionary storing information about estimated energies (corrected)
    #     raise NotImplementedError

    @abc.abstractmethod
    def estimate_energy(self,
                        experiment_key: str) -> float:
        raise NotImplementedError


    #
    # @abc.abstractmethod
    # def estimate_corrected_energy(self,
    #                               experiment_key: str
    #                               ) -> float:
    #     raise NotImplementedError



import sys
import os
import random
import pickle 

from tqdm import tqdm
import numpy as np

import qrem.common.providers.ibmutils.data_converters


from qrem.common import io 
import qrem.functions_qrem.functions_data_analysis as fdt 

from qrem.noise_characterization.tomography_design.overlapping.DOTMarginalsAnalyzer import DOTMarginalsAnalyzer 
from qrem.noise_characterization.tomography_design.overlapping.QDTMarginalsAnalyzer import QDTMarginalsAnalyzer

from qrem.noise_characterization.base_classes.OverlappingTomographyBase import OverlappingTomographyBase
from qrem.noise_characterization.tomography_design.overlapping.SeparableCircuitsCreator import SeparableCircuitsCreator
from qrem.noise_characterization.data_analysis.InitialNoiseAnalyzer import InitialNoiseAnalyzer
from typing import Dict, Tuple, List


#from qrem.backends_support.qiskit import qiskit_utilities
from qrem.functions_qrem import functions_data_analysis as fdt

from qrem.noise_model_generation.CN.NoiseModelGenerator import NoiseModelGenerator
from qrem.functions_qrem import functions_coherence_analysis as fca

import qrem.common.math as qrem_math
from qrem.common.printer import qprint, qprint_array
from qrem.types import CNModelData
from qrem.cn import simulation as cnsimulation 
from datetime import date
from typing import Tuple, Dict, List, Optional
import copy
import time
import re
from qrem.common.printer import qprint


import statistics 

def change_dictionary_format(noise_matrix:Dict) -> Dict:

    new_format ={}
    
    for key, matrix in noise_matrix.items():

        if key != 'averaged':
            
        
                   
            new_index = [int(character) for character in key]
        
            new_format[tuple(new_index)] = matrix
        
        else:

            new_format[key] = matrix

    return new_format


def change_state_dependent_noise_matrix_format(noise_matrix:Dict) -> Dict:

    """
    Function transforming old to new format of state dependent matrix 

    """

    state_dependent_noise_matrix_in_new_format = {}


    for key in noise_matrix.keys():
        
        if key != 'averaged':
            
        
            for neighbors_state, state_noise_matrix in noise_matrix[key].items():
        
                new_index = [int(character) for character in neighbors_state]
        
                state_dependent_noise_matrix_in_new_format[tuple(new_index)] = state_noise_matrix
            
        elif key == 'averaged':

            state_dependent_noise_matrix_in_new_format[key] = noise_matrix[key]
    
    return(state_dependent_noise_matrix_in_new_format)
    


def change_state_dependent_noise_matrices_dictionary_format(noise_matrices_dictionary:Dict) -> Dict:

    """
    Function transforming old to new format of state dependent matrix 

    """

    state_dependent_noise_matrices_dictionary_in_new_format = {}


    for key, item in noise_matrices_dictionary.items():

        state_dependent_noise_matrices_dictionary_in_new_format[key] = change_state_dependent_noise_matrix_format(noise_matrix=item)
        
        
    
    return(state_dependent_noise_matrices_dictionary_in_new_format)

def change_format_of_cn_dictionary(cn_dictionary:Dict)->Dict:

    pattern='q[0-9]+'
    cn_dictionary_new = {}
    for key,entry in cn_dictionary.items():
        index_list = re.findall(pattern,key)

        cluster_arrangement = []
        for index in index_list:
            cluster_arrangement.append(int(index[1:]))
        if entry == []:
            new_key = None
        else:
            new_key = tuple(entry)
        cn_dictionary_new[tuple(cluster_arrangement)] = new_key
    return cn_dictionary_new

def divide_data_into_characterization_benchmark_coherence_witness_data(results_dictionary:Dict[str,Dict[str,int]],marginals_dictionary:Dict[str,Dict[str,int]],ground_states_list=None, coherence_witnesses_list = None ):

    characterization_data_dictionary = {}
    
    
    if ground_states_list != None:
        benchmarks_results_dictionary = {}
        benchmarks_marginals_dictionary = {}
        characterization_results_dictionary = copy.copy(results_dictionary)
        characterization_marginals_dictionary = copy.copy(marginals_dictionary)
        
        for state in ground_states_list:
            benchmarks_results_dictionary[state] = copy.copy(results_dictionary[state])
            benchmarks_marginals_dictionary[state] =copy.copy( marginals_dictionary[state])
            del characterization_results_dictionary[state]
            del characterization_marginals_dictionary[state]
        
        characterization_data_dictionary['benchmarks_results_dictionary'] = benchmarks_results_dictionary
        characterization_data_dictionary['benchmarks_marginals_dictionary'] = benchmarks_marginals_dictionary 

    else:
        characterization_results_dictionary = results_dictionary
        characterization_marginals_dictionary = marginals_dictionary
        benchmarks_marginals_dictionary  = {}
        benchmarks_results_dictionary = {}
    
    if coherence_witnesses_list != None:
        coherence_witness_results_dictionary = {}
        coherence_witness_marginals_dictionary = {}
        for setting in coherence_witnesses_list:
            setting_string = ''
            
            for element in setting:
                setting_string = setting_string+ str(element)
            coherence_witness_results_dictionary[setting_string] = copy.copy(results_dictionary[setting_string])
            coherence_witness_marginals_dictionary[setting_string] =copy.copy( marginals_dictionary[setting_string])
            del characterization_results_dictionary[setting_string]
            del characterization_marginals_dictionary[setting_string]
        characterization_data_dictionary['coherence_witnesses_results_dictionary'] = coherence_witness_results_dictionary
        characterization_data_dictionary['coherence_witnesses_marginals_dictionary'] = coherence_witness_marginals_dictionary
    
    characterization_data_dictionary['characterization_results_dictionary'] = characterization_results_dictionary
    characterization_data_dictionary['characterization_marginals_dictionary'] = characterization_marginals_dictionary

    return characterization_data_dictionary 

def compute_marginals_dictionary(results_dictionary:Dict[str,Dict[str,int]],marginals_list:List,experiment_type:str='DDOT', data_directory:str ='', name_string=''):

    if experiment_type == 'DDOT':
        marginals_analyzer = DOTMarginalsAnalyzer(results_dictionary_ddot=results_dictionary)
        marginals_analyzer.compute_all_marginals(subsets_dictionary=marginals_list,show_progress_bar=True,multiprocessing=True)
        
        
            #marginals  calculation for QDOT
    elif experiment_type == 'QDOT':
        marginals_analyzer = QDTMarginalsAnalyzer(results_dictionary,experiment_name='QDT')
        marginals_analyzer.initialize_labels_interpreter(interpreter='PAULI')
        marginals_analyzer.compute_all_marginals(subsets_dictionary=marginals_list,show_progress_bar=True,multiprocessing=True)
        
    marginals_dictionary=marginals_analyzer.marginals_dictionary
  
    
    return marginals_dictionary

def compute_reduced_POVMs_and_noise_matrices(results_dictionary:Dict[str,Dict[str,int]],marginals_dictionary:Dict[str,Dict[str,int]] ,subset_of_qubits:List,experiment_type:str='DDOT'):

    if experiment_type == 'DDOT':
        marginals_analyzer = DOTMarginalsAnalyzer(results_dictionary_ddot=results_dictionary,marginals_dictionary=marginals_dictionary)


        marginals_analyzer.compute_subset_noise_matrices_averaged(subset_of_qubits)

        POVMs_dictionary_diag= marginals_analyzer._noise_matrices_dictionary

        #POVMs are established form noise matrices
        POVMs_dictionary = {}
        for key, entry in POVMs_dictionary_diag.items():
            list_of_matrices=[]
            for matrix in POVMs_dictionary_diag[key]['averaged']:
                list_of_matrices.append(np.diag(matrix))

            POVMs_dictionary[key]=list_of_matrices

      #marginals and POVM calculation for QDOT
    elif experiment_type == 'QDOT':
        marginals_analyzer = QDTMarginalsAnalyzer(results_dictionary=results_dictionary,experiment_name='QDT',marginals_dictionary=marginals_dictionary)

        marginals_analyzer.initialize_labels_interpreter(interpreter='PAULI')

        #calculate reduced POVMs
        marginals_analyzer.compute_subsets_POVMs_averaged(subsets_of_qubits=subset_of_qubits,
                                                           show_progress_bar=True,
                                                           estimation_method='PLS')
        POVMs_dictionary = marginals_analyzer._POVM_dictionary

        marginals_analyzer.compute_noise_matrices_from_POVMs(subsets_of_qubits=subset_of_qubits,
                                                     show_progress_bar=True)

    noise_matrices_dictionary = marginals_analyzer._noise_matrices_dictionary

    return POVMs_dictionary, noise_matrices_dictionary

def perform_clustering_routine(results_dictionary:Dict[str,Dict[str,int]],marginals_dictionary:Dict[str,Dict[str,int]], noise_matrices_dictionary:Dict[str,Dict[str,np.array]],correlations_data:np.array, number_of_qubits:int,clustering_functions_parameters: Dict ,perform_neighbors_search:bool = False):

    noise_model_analyzer = NoiseModelGenerator(results_dictionary=results_dictionary,
                                           marginals_dictionary=marginals_dictionary,
                                           noise_matrices_dictionary=noise_matrices_dictionary,
                                           correlations_data=correlations_data
                                           )

    all_clusters_sets_dictionary = {tuple([(qi,) for qi in range(number_of_qubits)]):None}

    #clustering starts 

    
    sizes_clusters = clustering_functions_parameters['sizes_clusters']

    distance_type = clustering_functions_parameters['distance_type']
    
    correlations_type = clustering_functions_parameters['correlations_type']

    alpha_hyperparameters = clustering_functions_parameters['alpha_hyperparameters'] 
    
    correlations_table = correlations_data[distance_type][correlations_type]

    all_clusters_sets_neighbors_dictionary = {tuple((qi,) for qi in range(number_of_qubits)):{(qj,) :None for qj in range(number_of_qubits)}}

    
        #product noise model is added to clusters list
    for max_cluster_size in sizes_clusters:
        qprint("\nCurrent max cluster size:", max_cluster_size, 'red')

        for alpha_multiplier in alpha_hyperparameters:
                
            clustering_function_arguments = {'alpha_multiplier': alpha_multiplier}

            clusters_tuples_now, score = noise_model_analyzer.compute_clusters(correlations_table=correlations_table,
                                                                            maximal_size=max_cluster_size,
                                                                           method_kwargs=clustering_function_arguments)
                           
            if perform_neighbors_search:
                    
                qprint('find_all_neighborhoods starts')
                                
                neighbors = noise_model_analyzer.find_all_neighborhoods(maximal_size= max_cluster_size+2)

                qprint('find_all_neighborhoods ends')
               
                        
            if clusters_tuples_now in all_clusters_sets_dictionary.keys():
                    
                all_clusters_sets_dictionary[clusters_tuples_now].append(
                        (max_cluster_size, clustering_function_arguments['alpha_multiplier'], score))
            else:
                
                all_clusters_sets_dictionary[clusters_tuples_now] = [
                        (max_cluster_size, clustering_function_arguments['alpha_multiplier'], score)]
                
            if perform_neighbors_search:
                    
                all_clusters_sets_neighbors_dictionary[clusters_tuples_now] = neighbors
            
            else:
                all_clusters_sets_neighbors_dictionary[clusters_tuples_now] = {cluster_tuple: None for cluster_tuple in clusters_tuples_now}



    

    return all_clusters_sets_neighbors_dictionary


def perform_noise_model_reconstruction_routine(results_dictionary:Dict[str,Dict[str,int]],number_of_qubits:int,all_clusters_sets_neighbors_dictionary:Dict,experiment_type:str = 'DDOT',find_neighbors:bool =False):

        
    noise_model_list =[]

    all_clusters_sets_list = list(all_clusters_sets_neighbors_dictionary.keys())
        
    for cluster_list in all_clusters_sets_list:

        if experiment_type =='DDOT':
                
            marginals_analyzer_noise_model = DOTMarginalsAnalyzer(results_dictionary_ddot=results_dictionary)

            new_noise_matrix_dictionary = {}

            cn_arrangement = all_clusters_sets_neighbors_dictionary[tuple(cluster_list)]

            if find_neighbors:
                
                cn_arrangement=change_format_of_cn_dictionary(cn_dictionary=cn_arrangement)



            for subset in cluster_list:

                if cn_arrangement[subset]  == None:

                    new_noise_matrix_dictionary[subset] =  {'averaged':marginals_analyzer_noise_model._compute_noise_matrix_averaged(subset=subset)}


                else:

                    new_noise_matrix_dictionary[subset] =  marginals_analyzer_noise_model._compute_noise_matrix_dependent(qubits_of_interest=subset,neighbors_of_interest=cn_arrangement[subset]) 

                        
                    new_noise_matrix_dictionary[subset]['averaged'] =  marginals_analyzer_noise_model._compute_noise_matrix_averaged(subset=subset)
                        



            marginals_analyzer_noise_model.compute_subset_noise_matrices_averaged(subsets_list=cluster_list,show_progress_bar=True)
                
            noise_matrices_dictionary = marginals_analyzer_noise_model.noise_matrices_dictionary

               
        elif experiment_type =='QDOT':
                
            marginals_analyzer_noise_model  = QDTMarginalsAnalyzer(results_dictionary=results_dictionary,experiment_name='QDT')
                
            marginals_analyzer_noise_model.initialize_labels_interpreter(interpreter='PAULI')

            #calculate reduced POVMs
            marginals_analyzer_noise_model.compute_subsets_POVMs_averaged(subsets_of_qubits=cluster_list, #cluster_substets
                                                            show_progress_bar=True,
                                                            estimation_method='PLS')
                
            POVMs_dictionary = marginals_analyzer_noise_model ._POVM_dictionary

            marginals_analyzer_noise_model .compute_noise_matrices_from_POVMs(subsets_of_qubits=cluster_list,
                                                        show_progress_bar=True)
                
            noise_matrices_dictionary = marginals_analyzer_noise_model.noise_matrices_dictionary

            new_noise_matrix_dictionary = {}

            for key, noise_matrix in noise_matrices_dictionary.items():
                    
                if noise_matrix != {}:
                        
                    new_noise_matrix_dictionary[key] =noise_matrix['averaged']

            #for test purposes

        

        noise_model = CNModelData(number_of_qubits)

        noise_model.set_clusters_tuple(clusters_tuple=tuple(cluster_list))

        noise_model.clusters_neighborhoods = cn_arrangement

        noise_model.set_noise_matrices_dictionary(noise_matrices_dictionary=new_noise_matrix_dictionary)

       

        for cluster, noise_matrix in noise_model.noise_matrices.items():

            noise_model.noise_matrices[cluster] = change_dictionary_format(noise_matrix=noise_matrix)

            
        noise_model_list.append(noise_model)

    return noise_model_list


def perform_POVMs_errors_and_correlation_coefficients_computation(results_dictionary:Dict,marginals_dictionary:Dict,POVMs_dictionary:Dict,distances_types:List = [('worst_case','classical')] ):

    noise_analyzer = InitialNoiseAnalyzer(results_dictionary=results_dictionary,
                                      marginals_dictionary=marginals_dictionary,
                                      POVM_dictionary=POVMs_dictionary
                                        )
    
    number_of_qubits = len(next(iter(results_dictionary.keys()))) 

    qubits_subset =[(qi,) for qi in range(number_of_qubits)]
    
    noise_analyzer.compute_errors_POVMs(qubits_subsets=qubits_subset,distances_types=distances_types)

    distances_types_correlations = [('worst_case', 'classical')]#,('worst_case', 'quantum'),('average case', 'classical'),('average case', 'quantum') ]

       
    
    noise_analyzer.compute_correlations_data_pairs(qubit_indices=range(number_of_qubits),
                                              distances_types=distances_types_correlations)

    correlations_data = noise_analyzer.correlations_data

    return {'POVMs_error_dictionary': noise_analyzer._errors_data_POVMs, 'correlations_data_dictionary' : correlations_data}






    









def execute_characterization_workflow(results_dictionary:Dict[str, Dict[str, int]], marginals_dictionary:Dict[str,Dict[Tuple, np.ndarray ]] = None, experiment_type: str = 'DDOT',   data_directory: str ='',number_of_benchmark_circuits: int = 0  , return_old_mitigation_data: bool = False, perform_clustering: bool =True, ground_states_list = None, coherence_witnesses_list = None,  name_id ='', perform_noise_model_reconstruction = True, find_neighbors = False):

    """
    A function that performs analysis of experimental data and provides various characteristics of readout noise as well as reconstructs a CN redout noise model. It returns a list of reconstructed CNModelData objects with noise models data.
    In addition, it saves data about correlation coefficients and reduced POVMs into files.    

    
    It is used to:

    1) Perform full characterization workflow
    2) As an initial step before running mitigation routine 


    Parameters
    ----------
    results_dictionary: dictionary (key - string encoding experimental setting , value - dictionary with experimental results i.e. key - string encoding outcome bits, values- number of occurrences of a given outcome  )
        Stores results of a characterization experiment. 

    marginals_dictionary: dictionary (key - string encoding experimental setting , value - dictionary with experimental results i.e. key - tuple encoding qubits in a marginal values- marginal probability distribution for a given marginal  ) 
        Optional parameter that allows to input precomputed marginal probability distributions into characterization routine. Stores marginal probability distributions for marginals. 

    experiment_type: string
        String encoding type of characterization experiment. Possible values: Diagonal detector overlapping tomography -  'DDOT' Quantum detector overlapping tomography - 'QDOT' 

    data_directory: string
        A string specifying path where results of characterization procedure are saved. 

    number_of_benchmark_circuits : integer
        The number of experiments from results_dictionary to be used in benchmarks. For number_of_benchmark_circuits > 0 the corresponding cumber of entries from results_dictionary is excluded from characterization procedure.  

    return_old_mitigation_data: bool
        Specifies whether old mitigation data are returned. Probably it should be excluded from the final release. It stays for now for debugging purposes. 
    
    perform_clustering: bool
       To be removed in final release: Specifies whether clustering algorithms to find noise model 
        

    
    Returns
    -------
    characterization_data_dictionary: dictionary
        A dictionary with results of characterization procedure
        
        Structure:
            Key: 'noise_models_list' - a list of noise models that is a result of clustering 
            
            Key: 'benchmarks_results_dictionary' - a dictionary with results to be used for benchmarks, empty unless number_of_benchmark_circuits > 0
            
            Key: 'benchmarks_marginals_dictionary' - a dictionary with marginals to be used for benchmarks, empty unless number_of_benchmark_circuits > 0
            
            Key: 'POVMs_dictionary' - a dictionary with two qubit POVMs
            
            Key: 'correlations_data' - a dictionary with correlation coefficients

            Old mitigation data:
            the rest of the keys is used to perform "old mitigation", it should be probably removed before final release, here it is used for debugging purposes 



    
        a value in a string

    Raises
    ------
    KeyError
        when a key error
    OtherError
        when an other error
    """    

    #number of qubits is read from results dictionary

    number_of_qubits = len(next(iter(results_dictionary.keys())))
    
 
    

    #this will be used in characterization functions
    qubit_indices = [i for i in range(number_of_qubits)]

    #here subsets of qubits used in initial caracterization are established. By default we perform analysis for single and two qubit substes.
    single_qubits = [(i,) for i in range(number_of_qubits)]
    pairs_of_qubits = [(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)]
    marginals_to_mitigate= single_qubits + pairs_of_qubits


    #this we use to save the results
    name_string = str(date.today()) + " q" + str(number_of_qubits)+name_id

    
   
  


    
    #computation of marginals is performed, if they are not provided
    if marginals_dictionary == None:
        
        marginals_dictionary = compute_marginals_dictionary(results_dictionary=results_dictionary,marginals_list=marginals_to_mitigate,experiment_type=experiment_type,data_directory=data_directory,name_string=name_string)
          
        qprint('MARGINALS COMPUTATION FINISHED')

        dictionary_to_save = {'marginals_dictionary':marginals_dictionary}
        file_name_marginals = experiment_type+ '_marginals_workflow_' + name_string +'.pkl'
        #io.save(dictionary_to_save=dictionary_to_save,
        #                        directory=data_directory,
        #                        custom_filename=file_name_marginals)
        
   

    characterization_data_dictionary = divide_data_into_characterization_benchmark_coherence_witness_data(results_dictionary=results_dictionary,marginals_dictionary=marginals_dictionary,ground_states_list=ground_states_list,coherence_witnesses_list=coherence_witnesses_list)

    characterization_marginals_dictionary = characterization_data_dictionary['characterization_marginals_dictionary']
    
    characterization_results_dictionary = characterization_data_dictionary['characterization_results_dictionary']

    benchmarks_results_dictionary = characterization_data_dictionary['benchmarks_results_dictionary']
    
    benchmarks_marginals_dictionary = characterization_data_dictionary['benchmarks_marginals_dictionary']  

    #noise matrices are computes, new object is created to prevent data to be overwritten 
    all_two_qubits_subsets = qrem_math.get_k_local_subsets(number_of_qubits, 2,True)
 
    characterization_POVMs_dictionary, characterization_matrices_dictionary = compute_reduced_POVMs_and_noise_matrices(results_dictionary=characterization_results_dictionary,marginals_dictionary=characterization_marginals_dictionary,subset_of_qubits=all_two_qubits_subsets,experiment_type='DDOT')
    
    qprint('REDUCED POVMS COMPUTATION FINISHED')    
   
    dictionary_to_save = {'POVMs_dictionary':characterization_POVMs_dictionary}

    file_name_POVMs  = 'DDOT_POVMs_PLS_workflow_' + name_string +'.pkl'

    #io.save(dictionary_to_save=dictionary_to_save,
    #                            directory=data_directory,
    #                            custom_filename=file_name_POVMs)
    
  

    noise_analysis_dictionary =  perform_POVMs_errors_and_correlation_coefficients_computation(results_dictionary=characterization_results_dictionary,marginals_dictionary=characterization_marginals_dictionary,POVMs_dictionary=characterization_POVMs_dictionary) 

    correlations_data = noise_analysis_dictionary['correlations_data_dictionary']

    POVMs_distances_dictionary = noise_analysis_dictionary['POVMs_error_dictionary']


   


   
    if name_id == 'IBM_Cusco':
        settings_list=['2','3','4','5']

    elif name_id == 'Rigetti_Aspen-M-3':
        settings_list=['2','3']
        
    coherence_bound_dictionary = fca.compute_coherence_indicator(marginals_dictionary=characterization_data_dictionary['coherence_witnesses_marginals_dictionary'],subset_list=[(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)],settings_list=settings_list)



 

    
      
    

    #here we compute only worst case classical distance, others can be added
   

    dictionary_to_save = {'correlations_data':correlations_data,
                      
                     }

    file_name_errors  = 'DDOT_correlations_workflow_' + name_string +'.pkl'
    #io.save(dictionary_to_save=dictionary_to_save,
    #                            directory=data_directory,
    #                            custom_filename=file_name_errors)



    qprint('CORRELATIONS COEFFICIENTS COMPUTATION FINISHED')  

    noise_model_analyzer = NoiseModelGenerator(results_dictionary=characterization_results_dictionary,
                                           marginals_dictionary=characterization_marginals_dictionary,
                                           noise_matrices_dictionary=characterization_matrices_dictionary,
                                           correlations_data=correlations_data
                                           )

    all_clusters_sets_dictionary = {tuple([(qi,) for qi in range(number_of_qubits)]):None}

    #clustering starts 

    clustering_functions_parameters = {'sizes_clusters':[2,3,4],'distance_type':'worst_case','correlations_type': 'classical','alpha_hyperparameters': [0] }
    
    all_clusters_sets_neighbors_dictionary = perform_clustering_routine(results_dictionary = characterization_results_dictionary,marginals_dictionary = characterization_marginals_dictionary, noise_matrices_dictionary = characterization_matrices_dictionary ,correlations_data = correlations_data, number_of_qubits=number_of_qubits,clustering_functions_parameters = clustering_functions_parameters)
    
    

    #########################################################################################################
    ###### Characterization routine ends here                                     ###########################  
    ###### What happens next is preparation of data used in mitigation routine    ###########################
    #########################################################################################################    


    
    all_clusters_sets_list = list(all_clusters_sets_neighbors_dictionary.keys())
    
    cluster_subsets = []

    for cluster_list in all_clusters_sets_list:
        for cluster in cluster_list:
            if cluster not in cluster_subsets:
                cluster_subsets.append(cluster)

    #marginals_analyzer.compute_all_marginals(cluster_subsets,
    #                                     show_progress_bar=True,
    #                                     multiprocessing=True)


    
    #all_clusters_sets_list = all_clusters_sets_list[1:]
    
    if perform_noise_model_reconstruction:
        #noise models data are saved in a list of CNModelData 
        
      
        noise_model_list = perform_noise_model_reconstruction_routine(results_dictionary=characterization_results_dictionary,number_of_qubits=number_of_qubits,all_clusters_sets_neighbors_dictionary=all_clusters_sets_neighbors_dictionary)

        characterization_data_dictionary['noise_models_list'] = noise_model_list
    
    characterization_data_dictionary['benchmarks_results_dictionary'] = benchmarks_results_dictionary
    characterization_data_dictionary['benchmarks_marginals_dictionary']= benchmarks_marginals_dictionary
    characterization_data_dictionary['POVMs_dictionary'] = characterization_POVMs_dictionary
    characterization_data_dictionary['correlations_data'] = correlations_data
    

    #old mitigation data can be also computed
    if return_old_mitigation_data:

        if experiment_type =='DDOT':
            marginals_analyzer_noise_model_old = DOTMarginalsAnalyzer(results_dictionary_ddot=characterization_results_dictionary)


            marginals_analyzer_noise_model_old.compute_subset_noise_matrices_averaged(subsets_list=cluster_subsets,show_progress_bar=True)
            noise_matrices_dictionary_old = marginals_analyzer_noise_model_old.noise_matrices_dictionary

        elif experiment_type =='QDOT':
            marginals_analyzer_noise_model_old  = QDTMarginalsAnalyzer(results_dictionary=characterization_results_dictionary,experiment_name='QDT')
            marginals_analyzer_noise_model_old .initialize_labels_interpreter(interpreter='PAULI')

        #calculate reduced POVMs
            marginals_analyzer_noise_model_old .compute_subsets_POVMs_averaged(subsets_of_qubits=cluster_subsets,
                                                           show_progress_bar=True,
                                                           estimation_method='PLS')
            POVMs_dictionary = marginals_analyzer_noise_model_old ._POVM_dictionary

            marginals_analyzer_noise_model_old .compute_noise_matrices_from_POVMs(subsets_of_qubits=cluster_subsets,
                                                     show_progress_bar=True)
            
                 
            new_noise_matrices_dictionary_old = marginals_analyzer_noise_model_old .noise_matrices_dictionary

            noise_matrices_dictionary_old = {}

            for key, noise_matrix in new_noise_matrices_dictionary_old .items():
                if noise_matrix != {}:
                    noise_matrices_dictionary_old[key] =noise_matrix

        
        correction_matrices, correction_indices,mitigation_data_dictionary = fdt.get_multiple_mitigation_strategies_clusters_for_pairs_of_qubits(
       
        pairs_of_qubits=pairs_of_qubits,
       
        clusters_sets=all_clusters_sets_dictionary,
       
        dictionary_results=characterization_results_dictionary,
       
        noise_matrices_dictionary=noise_matrices_dictionary_old,
        show_progress_bar=True)

        characterization_data_dictionary['correction_matrices'] = correction_matrices

        characterization_data_dictionary['correlation_coefficients_dictionary'] = correlations_data
        
        characterization_data_dictionary['POVMs_dictionary']= characterization_POVMs_dictionary
        
        characterization_data_dictionary['POVMs_distances_dictionary']=POVMs_distances_dictionary

        
        characterization_data_dictionary['correction_indices'] = correction_indices
        
        characterization_data_dictionary['mitigation_data_dictionary'] = mitigation_data_dictionary
        
        characterization_data_dictionary['all_clusters_sets_dictionary'] = all_clusters_sets_dictionary
        
        characterization_data_dictionary['noise_matrices_dictionary_old'] = noise_matrices_dictionary_old
        
        characterization_data_dictionary['all_clusters_sets_neighbors_dictionary'] = all_clusters_sets_neighbors_dictionary

        characterization_data_dictionary['coherence_bound_dictionary']  = coherence_bound_dictionary




        
       
        
    file_name_mitigation_data  = 'DDOT_characterization_data_' + name_string +'.pkl'
    #io.save(dictionary_to_save=characterization_data_dictionary,
    #                            directory=data_directory,
    #                            custom_filename=file_name_mitigation_data)

    
    return characterization_data_dictionary
    


    





def simulate_clean_experiment():
    """1. przenieść src\\qrem\\usecases\\mitigation\\create_mitigation_data_KKM.py funkcja: simulate_experiment do  
    src\\qrem\\common\\providers\\simulation.py->simulate_clean_experiment()
    """
    pass


def simulate_noisy_experiment():
    """2.stworzyć funkcję simulate_noisy_experiment() w \\src\\qrem\\common\\providers\\simulation.py, która na podstawie wyników simulate_clean_experiment i parametru noise_model="CN" i innych zwróci noisy_results, czyli np. w przypadku CN wywoa cunkcję:
	src\\qrem\\cn\\simulation.py::simulate_noise_results_dictionary()
    """

    pass